"""
Unit tests for the recorder functions (_recorder.py module).
"""

from __future__ import absolute_import, print_function

# Allows use of lots of single character variable names.
# pylint: disable=invalid-name,missing-docstring,too-many-statements
# pylint: disable=too-many-lines,no-self-use
import sys
import os
import os.path
import logging
import logging.handlers
import warnings
from io import open as _open
from datetime import datetime, timedelta

import pytest
import six
from testfixtures import LogCapture
# Enabled only to display a tree of loggers
# from logging_tree import printout as logging_tree_printout
import yaml
import yamlloader
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from ...utils import skip_if_moftab_regenerated, is_inherited_from
from ..utils.dmtf_mof_schema_def import install_test_dmtf_schema
from ..utils.pytest_extensions import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ...utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import CIMInstanceName, CIMInstance, CIMClassName, CIMClass, \
    CIMProperty, CIMMethod, CIMParameter, CIMQualifier, \
    CIMQualifierDeclaration, \
    Uint8, Uint16, Uint32, Uint64, Sint8, Sint16, \
    Sint32, Sint64, Real32, Real64, CIMDateTime, MinutesFromUTC, CIMError, \
    HTTPError, WBEMConnection, LogOperationRecorder, BaseOperationRecorder, \
    configure_logger  # noqa: E402
# Renamed the following import to not have py.test pick it up as a test class:
from pywbem import TestClientRecorder as _TestClientRecorder  # noqa: E402
from pywbem._cim_operations import pull_path_result_tuple, \
    pull_inst_result_tuple  # noqa: E402
from pywbem._utils import _format  # noqa: E402
pywbem_mock = import_installed('pywbem_mock')
from pywbem_mock import FakedWBEMConnection  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name

# Ordered dict type created by yamlloader.ordereddict loaders
if sys.version_info[0:2] >= (3, 7):
    yaml_ordereddict = dict  # pylint: disable=invalid-name
else:
    yaml_ordereddict = OrderedDict  # pylint: disable=invalid-name

TEST_DIR = os.path.dirname(__file__)

# test outpuf file for the recorder tests.  This is opened for each
# test to save yaml output and may be reloaded during the same test
# to confirm the yaml results.
TEST_YAML_FILE = os.path.join(TEST_DIR, 'test_recorder.yaml')

TEST_OUTPUT_LOG = os.path.join(TEST_DIR, 'test_recorder.log')

VERBOSE = False
DEBUG_TEST_YAML_FILE = False  # Show the generated test client YAML file


# CIMProperty objects of all types and the corresponding test client YAML
CIMPROPERTY_B1_OBJ = CIMProperty('B1', value=True)
CIMPROPERTY_B1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'B1',
    type=u'boolean',
    value=True,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_C1_OBJ = CIMProperty('C1', type='char16', value=u'A')
CIMPROPERTY_C1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'C1',
    type=u'char16',
    value=u'A',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_C2_OBJ = CIMProperty('C2', type='char16', value=u'\u00E4')
CIMPROPERTY_C2_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'C2',
    type=u'char16',
    value=u'\u00E4',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
# Note that the value space of the char16 datatype is UCS-2
CIMPROPERTY_S1_OBJ = CIMProperty('S1', type='string', value=u'Ham')
CIMPROPERTY_S1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S1',
    type=u'string',
    value=u'Ham',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S2_OBJ = CIMProperty('S2', type='string', value=u'H\u00E4m')
CIMPROPERTY_S2_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S2',
    type=u'string',
    value=u'H\u00E4m',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S3_OBJ = CIMProperty('S3', type='string', value=u'A\U00010142B')
CIMPROPERTY_S3_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S3',
    type=u'string',
    value=u'A\U00010142B',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_E1_OBJ = CIMProperty('E1', type='string',
                                 embedded_object='instance',
                                 value=CIMInstance('C_Emb'))
CIMPROPERTY_E1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'E1',
    type=u'string',
    value=dict(
        pywbem_object='CIMInstance',
        classname='C_Emb',
        properties=yaml_ordereddict(),
        qualifiers=yaml_ordereddict(),
        path=None,
    ),
    reference_class=None,
    embedded_object=u'instance',
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_E2_OBJ = CIMProperty('E2', type='string',
                                 embedded_object='object',
                                 value=CIMInstance('C_Emb'))
CIMPROPERTY_E2_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'E2',
    type=u'string',
    value=dict(
        pywbem_object='CIMInstance',
        classname='C_Emb',
        properties=yaml_ordereddict(),
        qualifiers=yaml_ordereddict(),
        path=None,
    ),
    reference_class=None,
    embedded_object=u'object',
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_E3_OBJ = CIMProperty('E3', type='string',
                                 embedded_object='object',
                                 value=CIMClass('C_Emb'))
CIMPROPERTY_E3_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'E3',
    type=u'string',
    value=dict(
        pywbem_object='CIMClass',
        classname='C_Emb',
        superclass=None,
        properties=yaml_ordereddict(),
        methods=yaml_ordereddict(),
        qualifiers=yaml_ordereddict(),
        path=None,
    ),
    reference_class=None,
    embedded_object=u'object',
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_U11_OBJ = CIMProperty('U11', value=Uint8(42))
CIMPROPERTY_U11_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'U11',
    type=u'uint8',
    value=42,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_U21_OBJ = CIMProperty('U21', value=Uint16(4216))
CIMPROPERTY_U21_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'U21',
    type=u'uint16',
    value=4216,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_U41_OBJ = CIMProperty('U41', value=Uint32(4232))
CIMPROPERTY_U41_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'U41',
    type=u'uint32',
    value=4232,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_U81_OBJ = CIMProperty('U81', value=Uint64(4264))
CIMPROPERTY_U81_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'U81',
    type=u'uint64',
    value=4264,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S11_OBJ = CIMProperty('S11', value=Sint8(-42))
CIMPROPERTY_S11_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S11',
    type=u'sint8',
    value=-42,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S21_OBJ = CIMProperty('S21', value=Sint16(-4216))
CIMPROPERTY_S21_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S21',
    type=u'sint16',
    value=-4216,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S41_OBJ = CIMProperty('S41', value=Sint32(-4232))
CIMPROPERTY_S41_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S41',
    type=u'sint32',
    value=-4232,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_S81_OBJ = CIMProperty('S81', value=Sint64(-4264))
CIMPROPERTY_S81_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'S81',
    type=u'sint64',
    value=-4264,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_R41_OBJ = CIMProperty('R41', value=Real32(42.0))
CIMPROPERTY_R41_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'R41',
    type=u'real32',
    value=42.0,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_R81_OBJ = CIMProperty('R81', value=Real64(42.64))
CIMPROPERTY_R81_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'R81',
    type=u'real64',
    value=42.64,
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_R1_OBJ = CIMProperty('R1', value=CIMInstanceName('C1'))
CIMPROPERTY_R1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'R1',
    type=u'reference',
    value=dict(
        pywbem_object='CIMInstanceName',
        classname='C1',
        keybindings=yaml_ordereddict(),
        namespace=None,
        host=None,
    ),
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)
CIMPROPERTY_D1_OBJ = CIMProperty(
    'D1',
    value=CIMDateTime('20140924193040.654321+120'))
CIMPROPERTY_D1_TCYAML = dict(
    pywbem_object='CIMProperty',
    name=u'D1',
    type=u'datetime',
    value=u'20140924193040.654321+120',
    reference_class=None,
    embedded_object=None,
    is_array=False,
    array_size=None,
    class_origin=None,
    propagated=None,
    qualifiers=yaml_ordereddict(),
)

# CIMInstance object without path, with all property types and the test client
# YAML object
CIMINSTANCE_ALL_OBJ = CIMInstance(
    'CIM_Foo',
    properties=[
        CIMPROPERTY_B1_OBJ,
        CIMPROPERTY_C1_OBJ,
        CIMPROPERTY_C2_OBJ,
        CIMPROPERTY_S1_OBJ,
        CIMPROPERTY_S2_OBJ,
        CIMPROPERTY_S3_OBJ,
        CIMPROPERTY_E1_OBJ,
        CIMPROPERTY_E2_OBJ,
        CIMPROPERTY_E3_OBJ,
        CIMPROPERTY_U11_OBJ,
        CIMPROPERTY_U21_OBJ,
        CIMPROPERTY_U41_OBJ,
        CIMPROPERTY_U81_OBJ,
        CIMPROPERTY_S11_OBJ,
        CIMPROPERTY_S21_OBJ,
        CIMPROPERTY_S41_OBJ,
        CIMPROPERTY_S81_OBJ,
        CIMPROPERTY_R41_OBJ,
        CIMPROPERTY_R81_OBJ,
        CIMPROPERTY_R1_OBJ,
        CIMPROPERTY_D1_OBJ,
    ],
    path=None,
)
CIMINSTANCE_ALL_TCYAML = yaml_ordereddict([
    ('pywbem_object', 'CIMInstance'),
    ('classname', 'CIM_Foo'),
    ('properties', yaml_ordereddict([
        ('B1', CIMPROPERTY_B1_TCYAML),
        ('C1', CIMPROPERTY_C1_TCYAML),
        ('C2', CIMPROPERTY_C2_TCYAML),
        ('S1', CIMPROPERTY_S1_TCYAML),
        ('S2', CIMPROPERTY_S2_TCYAML),
        ('S3', CIMPROPERTY_S3_TCYAML),
        ('E1', CIMPROPERTY_E1_TCYAML),
        ('E2', CIMPROPERTY_E2_TCYAML),
        ('E3', CIMPROPERTY_E3_TCYAML),
        ('U11', CIMPROPERTY_U11_TCYAML),
        ('U21', CIMPROPERTY_U21_TCYAML),
        ('U41', CIMPROPERTY_U41_TCYAML),
        ('U81', CIMPROPERTY_U81_TCYAML),
        ('S11', CIMPROPERTY_S11_TCYAML),
        ('S21', CIMPROPERTY_S21_TCYAML),
        ('S41', CIMPROPERTY_S41_TCYAML),
        ('S81', CIMPROPERTY_S81_TCYAML),
        ('R41', CIMPROPERTY_R41_TCYAML),
        ('R81', CIMPROPERTY_R81_TCYAML),
        ('R1', CIMPROPERTY_R1_TCYAML),
        ('D1', CIMPROPERTY_D1_TCYAML),
    ])),
    ('qualifiers', yaml_ordereddict()),
    ('path', None),
])


@pytest.fixture(params=[
    (CIMINSTANCE_ALL_OBJ, CIMINSTANCE_ALL_TCYAML),
], scope='module')
def instance_tuple(request):
    """
    Fixture for a CIMInstance object without path and its corresponding
    test client YAML object.

    Returns a tuple(CIMInstance, tcyaml_dict).
    """
    return request.param


# CIMInstance object with path, with all property types and the test client
# YAML object
CIMINSTANCE_WP_ALL_OBJ = CIMInstance(
    'CIM_Foo',
    properties=[
        CIMPROPERTY_B1_OBJ,
        CIMPROPERTY_C1_OBJ,
        CIMPROPERTY_C2_OBJ,
        CIMPROPERTY_S1_OBJ,
        CIMPROPERTY_S2_OBJ,
        CIMPROPERTY_S3_OBJ,
        CIMPROPERTY_E1_OBJ,
        CIMPROPERTY_E2_OBJ,
        CIMPROPERTY_E3_OBJ,
        CIMPROPERTY_U11_OBJ,
        CIMPROPERTY_U21_OBJ,
        CIMPROPERTY_U41_OBJ,
        CIMPROPERTY_U81_OBJ,
        CIMPROPERTY_S11_OBJ,
        CIMPROPERTY_S21_OBJ,
        CIMPROPERTY_S41_OBJ,
        CIMPROPERTY_S81_OBJ,
        CIMPROPERTY_R41_OBJ,
        CIMPROPERTY_R81_OBJ,
        CIMPROPERTY_R1_OBJ,
        CIMPROPERTY_D1_OBJ,
    ],
    path=CIMInstanceName(
        'CIM_Foo',
        keybindings=dict(S1=CIMPROPERTY_S1_OBJ.value),
        namespace='root/cimv2',
        host='woot.com',
    ),
)
CIMINSTANCE_WP_ALL_TCYAML = yaml_ordereddict([
    ('pywbem_object', 'CIMInstance'),
    ('classname', 'CIM_Foo'),
    ('properties', yaml_ordereddict([
        ('B1', CIMPROPERTY_B1_TCYAML),
        ('C1', CIMPROPERTY_C1_TCYAML),
        ('C2', CIMPROPERTY_C2_TCYAML),
        ('S1', CIMPROPERTY_S1_TCYAML),
        ('S2', CIMPROPERTY_S2_TCYAML),
        ('S3', CIMPROPERTY_S3_TCYAML),
        ('E1', CIMPROPERTY_E1_TCYAML),
        ('E2', CIMPROPERTY_E2_TCYAML),
        ('E3', CIMPROPERTY_E3_TCYAML),
        ('U11', CIMPROPERTY_U11_TCYAML),
        ('U21', CIMPROPERTY_U21_TCYAML),
        ('U41', CIMPROPERTY_U41_TCYAML),
        ('U81', CIMPROPERTY_U81_TCYAML),
        ('S11', CIMPROPERTY_S11_TCYAML),
        ('S21', CIMPROPERTY_S21_TCYAML),
        ('S41', CIMPROPERTY_S41_TCYAML),
        ('S81', CIMPROPERTY_S81_TCYAML),
        ('R41', CIMPROPERTY_R41_TCYAML),
        ('R81', CIMPROPERTY_R81_TCYAML),
        ('R1', CIMPROPERTY_R1_TCYAML),
        ('D1', CIMPROPERTY_D1_TCYAML),
    ])),
    ('qualifiers', yaml_ordereddict()),
    ('path', yaml_ordereddict([
        ('pywbem_object', 'CIMInstanceName'),
        ('classname', 'CIM_Foo'),
        ('keybindings', dict(S1=CIMPROPERTY_S1_TCYAML['value'])),
        ('namespace', 'root/cimv2'),
        ('host', 'woot.com'),
    ])),
])


@pytest.fixture(params=[
    (CIMINSTANCE_WP_ALL_OBJ, CIMINSTANCE_WP_ALL_TCYAML),
], scope='module')
def instance_wp_tuple(request):
    """
    Fixture for a CIMInstance object with path and its corresponding
    test client YAML object.

    Returns a tuple(CIMInstance, tcyaml_dict).
    """
    return request.param


# Two CIMInstance objects without path and the test client YAML object
CIMINSTANCES_OBJS = [
    CIMInstance(
        'CIM_Foo',
        properties=[
            CIMPROPERTY_S1_OBJ,
            CIMPROPERTY_S2_OBJ,
        ],
        path=None,
    ),
    CIMInstance(
        'CIM_Foo',
        properties=[
            CIMPROPERTY_S3_OBJ,
        ],
        path=None,
    ),
]
CIMINSTANCES_TCYAML = [
    yaml_ordereddict([
        ('pywbem_object', 'CIMInstance'),
        ('classname', 'CIM_Foo'),
        ('properties', yaml_ordereddict([
            ('S1', CIMPROPERTY_S1_TCYAML),
            ('S2', CIMPROPERTY_S2_TCYAML),
        ])),
        ('qualifiers', yaml_ordereddict()),
        ('path', None),
    ]),
    yaml_ordereddict([
        ('pywbem_object', 'CIMInstance'),
        ('classname', 'CIM_Foo'),
        ('properties', yaml_ordereddict([
            ('S3', CIMPROPERTY_S3_TCYAML),
        ])),
        ('qualifiers', yaml_ordereddict()),
        ('path', None),
    ]),
]


@pytest.fixture(params=[
    (CIMINSTANCES_OBJS, CIMINSTANCES_TCYAML),
], scope='module')
def instances_tuple(request):
    """
    Fixture for a list of two CIMInstance objects without path and its
    corresponding test client YAML object.

    Returns a tuple(list(CIMInstance), tcyaml_dict).
    """
    return request.param


# Two CIMInstance objects with path and the test client YAML object
CIMINSTANCES_WP_OBJS = [
    CIMInstance(
        'CIM_Foo',
        properties=[
            CIMPROPERTY_S1_OBJ,
            CIMPROPERTY_S2_OBJ,
        ],
        path=CIMInstanceName(
            'CIM_Foo',
            keybindings=dict(S1=CIMPROPERTY_S1_OBJ.value),
            namespace='root/cimv2',
            host='woot.com',
        ),
    ),
    CIMInstance(
        'CIM_Foo',
        properties=[
            CIMPROPERTY_S1_OBJ,
            CIMPROPERTY_S3_OBJ,
        ],
        path=CIMInstanceName(
            'CIM_Foo',
            keybindings=dict(S1=CIMPROPERTY_S1_OBJ.value),
            namespace='root/cimv2',
            host='woot.com',
        ),
    ),
]
CIMINSTANCES_WP_TCYAML = [
    yaml_ordereddict([
        ('pywbem_object', 'CIMInstance'),
        ('classname', 'CIM_Foo'),
        ('properties', yaml_ordereddict([
            ('S1', CIMPROPERTY_S1_TCYAML),
            ('S2', CIMPROPERTY_S2_TCYAML),
        ])),
        ('qualifiers', yaml_ordereddict()),
        ('path', yaml_ordereddict([
            ('pywbem_object', 'CIMInstanceName'),
            ('classname', 'CIM_Foo'),
            ('keybindings', dict(S1=CIMPROPERTY_S1_TCYAML['value'])),
            ('namespace', 'root/cimv2'),
            ('host', 'woot.com'),
        ])),
    ]),
    yaml_ordereddict([
        ('pywbem_object', 'CIMInstance'),
        ('classname', 'CIM_Foo'),
        ('properties', yaml_ordereddict([
            ('S1', CIMPROPERTY_S1_TCYAML),
            ('S3', CIMPROPERTY_S3_TCYAML),
        ])),
        ('qualifiers', yaml_ordereddict()),
        ('path', yaml_ordereddict([
            ('pywbem_object', 'CIMInstanceName'),
            ('classname', 'CIM_Foo'),
            ('keybindings', dict(S1=CIMPROPERTY_S1_TCYAML['value'])),
            ('namespace', 'root/cimv2'),
            ('host', 'woot.com'),
        ])),
    ]),
]


@pytest.fixture(params=[
    (CIMINSTANCES_WP_OBJS, CIMINSTANCES_WP_TCYAML),
], scope='module')
def instances_wp_tuple(request):
    """
    Fixture for a list of two CIMInstance objects with path and its
    corresponding test client YAML object.

    Returns a tuple(list(CIMInstance), tcyaml_dict).
    """
    return request.param


# CIMInstanceName object and the test client YAML object
CIMINSTANCENAME_OBJ = CIMInstanceName(
    'CIM_Foo',
    keybindings=dict(S1='a'),
    namespace='root/cimv2',
    host='woot.com',
)
CIMINSTANCENAME_TCYAML = yaml_ordereddict([
    ('pywbem_object', 'CIMInstanceName'),
    ('classname', 'CIM_Foo'),
    ('keybindings', dict(S1='a')),
    ('namespace', 'root/cimv2'),
    ('host', 'woot.com'),
])


@pytest.fixture(params=[
    (CIMINSTANCENAME_OBJ, CIMINSTANCENAME_TCYAML),
], scope='module')
def instancename_tuple(request):
    """
    Fixture for a CIMInstanceName object and its corresponding
    test client YAML object.

    Returns a tuple(CIMInstanceName, tcyaml_dict).
    """
    return request.param


@pytest.fixture(autouse=True)
def capture():
    """
    Fixture for log capturing. Returns a testfixtures.LogCapture object.

    Note: The log_capture decorator of testfixtures is not compatible with
    pytest fixtures, see
    https://testfixtures.readthedocs.io/en/latest/logging.html#the-decorator
    """
    with LogCapture() as lc:
        yield lc


@pytest.fixture(scope='function')
def test_client_recorder(request):
    # pylint: disable=unused-argument
    """
    Fixture for a TestClientRecorder object that records to TEST_YAML_FILE and
    that is enabled.
    """
    fp = _TestClientRecorder.open_file(TEST_YAML_FILE, 'w')
    recorder = _TestClientRecorder(fp)
    recorder.reset()
    recorder.enable()
    # recorder.test_fp = fp
    yield recorder
    # del recorder.test_fp
    fp.close()


def logged_payload(payload):
    """
    Return the payload unicode string as it is logged by the pywbem recorders.
    """
    ret_payload = repr(payload)
    if ret_payload.startswith("u'"):
        ret_payload = ret_payload[1:]
    return ret_payload


def load_recorder_yaml_file():
    """
    Load the test YAML file created by the test recorder and return its content
    as a dict.
    """
    with _open(TEST_YAML_FILE, encoding="utf-8") as fp:
        yaml_content = yaml.load(
            fp, Loader=yamlloader.ordereddict.CSafeLoader)
    return yaml_content


def cleanup_recorder_yaml_file():
    """
    Delete the test YAML file created by the test recorder, to clean up.
    """
    if os.path.exists(TEST_YAML_FILE):
        try:
            os.remove(TEST_YAML_FILE)
        except Exception as exc:  # pylint: disable=broad-except
            # The file is still open at this point and on Windows,
            # a WindowsError is raised.
            warnings.warn(
                "Cleaning up the test client YAML file {} failed: {}".
                format(TEST_YAML_FILE, exc), UserWarning)


def test_BaseOperationRecorder_init():
    """
    Test function for BaseOperationRecorder.__init__()
    """

    # BaseOperationRecorder is an abstract base class, so we test its methods
    # through its derived class TestClientRecorder, but we verify that the
    # derived class attributes/methods we use were inherited from the base
    # class.
    assert is_inherited_from(
        'enabled', _TestClientRecorder, BaseOperationRecorder)

    try:
        fp = open(os.devnull, 'w')

        # The code to be tested
        recorder = _TestClientRecorder(fp)

        assert recorder.enabled is True

    finally:
        fp.close()


def test_BaseOperationRecorder_enable_disable():
    """
    Test function for BaseOperationRecorder.enable() and disable()
    """

    # BaseOperationRecorder is an abstract base class, so we test its methods
    # through its derived class TestClientRecorder, but we verify that the
    # derived class attributes/methods we use were inherited from the base
    # class.
    assert is_inherited_from(
        'disable', _TestClientRecorder, BaseOperationRecorder)
    assert is_inherited_from(
        'enable', _TestClientRecorder, BaseOperationRecorder)
    assert is_inherited_from(
        'enabled', _TestClientRecorder, BaseOperationRecorder)

    try:
        fp = open(os.devnull, 'w')

        recorder = _TestClientRecorder(fp)

        # The code to be tested
        recorder.enable()
        assert recorder.enabled is True

        # The code to be tested
        recorder.disable()
        assert recorder.enabled is False

        # The code to be tested
        recorder.disable()
        assert recorder.enabled is False

        # The code to be tested
        recorder.enable()
        assert recorder.enabled is True

    finally:
        fp.close()


TESTCASES_BASEOPERATIONRECORDER_OPEN_FILE = [

    # Testcases for BaseOperationRecorder.open_file()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * text: Input data for the file, as unicode string.
    #   * exp_bytes: Expected UTF-8 Bytes in the file, as byte string.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Single 7-bit ASCII character",
        dict(
            text=u'A',
            exp_bytes=b'A',
        ),
        None, None, True
    ),
    (
        "Single UCS-2 character resulting in 2-char UTF-8 sequence",
        dict(
            text=u'\u00E8',
            exp_bytes=b'\xC3\xA8',
        ),
        None, None, True
    ),
    (
        "Single UCS-2 character resulting in 3-char UTF-8 sequence",
        dict(
            text=u'\u2014',
            exp_bytes=b'\xE2\x80\x94',
        ),
        None, None, True
    ),
    (
        "Single UTF-16 surrogate sequence for U+10142 resulting in 4-char "
        "UTF-8 sequence",
        dict(
            text=u'\uD800\uDD42',
            exp_bytes=b'\xF0\x90\x85\x82',
        ),
        None, None, six.PY2  # Not supported in Python 3
    ),
    (
        "Single UCS-4 character resulting in 4-char UTF-8 sequence",
        dict(
            text=u'\U00010142',
            exp_bytes=b'\xF0\x90\x85\x82',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_BASEOPERATIONRECORDER_OPEN_FILE)
@simplified_test_function
def test_BaseOperationRecorder_open_file(testcase, text, exp_bytes):
    # pylint: disable=unused-argument
    """
    Test function for BaseOperationRecorder.open_file()
    """

    # BaseOperationRecorder is an abstract base class, so we test its methods
    # through its derived class TestClientRecorder, but we verify that the
    # derived class attributes/methods we use were inherited from the base
    # class.
    assert is_inherited_from(
        'open_file', _TestClientRecorder, BaseOperationRecorder)

    tmp_filename = 'openfile.tmp'

    # The code to be tested
    fp = _TestClientRecorder.open_file(tmp_filename, 'w')

    fp.write(text)
    fp.close()

    with open(tmp_filename, 'rb') as fp:
        act_bytes = fp.read()
    assert act_bytes == exp_bytes

    os.remove(tmp_filename)


TESTCASES_TESTCLIENTRECORDER_TOYAML = [

    # Testcases for TestClientRecorder.toyaml()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: Object that has dictionary behavior, e.g. IMInstanceName.
    #   * exp_yaml: Expected YAML object, for validation.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Boolean value True",
        dict(
            obj=True,
            exp_yaml=True,
        ),
        None, None, True
    ),
    (
        "Boolean value False",
        dict(
            obj=False,
            exp_yaml=False,
        ),
        None, None, True
    ),
    (
        "Integer value",
        dict(
            obj=1234,
            exp_yaml=1234,
        ),
        None, None, True
    ),
    (
        "Uint8 value",
        dict(
            obj=Uint8(42),
            exp_yaml=42,
        ),
        None, None, True
    ),
    (
        "Float value",
        dict(
            obj=42.1,
            exp_yaml=None,
        ),
        TypeError, None, True
    ),
    (
        "Real32 value",
        dict(
            obj=Real32(42.1),
            exp_yaml=42.1,
        ),
        None, None, True
    ),
    (
        "Unicode string value",
        dict(
            obj='blahblah ',
            exp_yaml='blahblah ',
        ),
        None, None, True
    ),
    (
        "Byte string value",
        dict(
            obj=b'blahblah ',
            exp_yaml='blahblah ',
        ),
        None, None, True
    ),
    (
        "CIMDateTime object",
        dict(
            obj=CIMDateTime('20140924193040.654321+120'),
            exp_yaml='20140924193040.654321+120',
        ),
        None, None, True
    ),
    (
        "datetime object",
        dict(
            obj=datetime(year=2020, month=1, day=28, hour=14, minute=46,
                         second=40, microsecond=654321,
                         tzinfo=MinutesFromUTC(120)),
            exp_yaml=CIMDateTime('20200128144640.654321+120'),
        ),
        None, None, True
    ),
    (
        "timedelta object",
        dict(
            obj=timedelta(183, (13 * 60 + 25) * 60 + 42, 234567),
            exp_yaml=CIMDateTime('00000183132542.234567:000'),
        ),
        None, None, True
    ),
    (
        "CIMInstanceName object",
        dict(
            obj=CIMInstanceName(
                'CIM_Foo',
                keybindings=OrderedDict([
                    ('Chicken', 'Ham'),
                    ('Beans', Uint8(42)),
                ]),
                namespace='cimv2',
                host='woot.com',
            ),
            exp_yaml=dict(
                pywbem_object='CIMInstanceName',
                classname=u'CIM_Foo',
                namespace=u'cimv2',
                host=u'woot.com',
                keybindings=yaml_ordereddict([
                    (u'Chicken', u'Ham'),
                    (u'Beans', 42),
                ]),
            ),
        ),
        None, None, True
    ),
    (
        "CIMInstance object",
        dict(
            obj=CIMInstance(
                'CIM_Foo',
                properties=[
                    ('Chicken', 'Ham'),
                ],
            ),
            exp_yaml=dict(
                pywbem_object='CIMInstance',
                classname=u'CIM_Foo',
                properties=yaml_ordereddict([
                    (u'Chicken', dict(
                        pywbem_object='CIMProperty',
                        name=u'Chicken',
                        type=u'string',
                        value=u'Ham',
                        reference_class=None,
                        embedded_object=None,
                        is_array=False,
                        array_size=None,
                        class_origin=None,
                        propagated=None,
                        qualifiers=yaml_ordereddict(),
                    )),
                ]),
                qualifiers=yaml_ordereddict(),
                path=None,
            ),
        ),
        None, None, True
    ),
    (
        "CIMClassName object",
        dict(
            obj=CIMClassName(
                'CIM_Foo',
                namespace='cimv2',
                host='woot.com',
            ),
            exp_yaml=dict(
                pywbem_object='CIMClassName',
                classname=u'CIM_Foo',
                namespace=u'cimv2',
                host='woot.com',
            ),
        ),
        None, None, True
    ),
    (
        "CIMClass object",
        dict(
            obj=CIMClass(
                'CIM_Foo',
                properties=[
                    CIMProperty('Chicken', type='string', value='Ham'),
                ],
            ),
            exp_yaml=dict(
                pywbem_object='CIMClass',
                classname=u'CIM_Foo',
                superclass=None,
                properties=yaml_ordereddict([
                    (u'Chicken', dict(
                        pywbem_object='CIMProperty',
                        name=u'Chicken',
                        type=u'string',
                        value=u'Ham',
                        reference_class=None,
                        embedded_object=None,
                        is_array=False,
                        array_size=None,
                        class_origin=None,
                        propagated=None,
                        qualifiers=yaml_ordereddict(),
                    )),
                ]),
                methods=yaml_ordereddict(),
                qualifiers=yaml_ordereddict(),
                path=None,
            ),
        ),
        None, None, True
    ),
    (
        "CIMProperty object",
        dict(
            obj=CIMProperty('Chicken', type='string', value='Ham'),
            exp_yaml=dict(
                pywbem_object='CIMProperty',
                name=u'Chicken',
                type=u'string',
                value=u'Ham',
                reference_class=None,
                embedded_object=None,
                is_array=False,
                array_size=None,
                class_origin=None,
                propagated=None,
                qualifiers=yaml_ordereddict(),
            ),
        ),
        None, None, True
    ),
    # TODO: Add testcase for reference property
    # TODO: Add testcase for embedded object property
    # TODO: Add testcase for array property
    (
        "CIMMethod object",
        dict(
            obj=CIMMethod('Chicken', return_type='string'),
            exp_yaml=dict(
                pywbem_object='CIMMethod',
                name=u'Chicken',
                return_type=u'string',
                class_origin=None,
                propagated=None,
                parameters=yaml_ordereddict(),
                qualifiers=yaml_ordereddict(),
            ),
        ),
        None, None, True
    ),
    (
        "CIMParameter object",
        dict(
            obj=CIMParameter('Chicken', type='string'),
            exp_yaml=dict(
                pywbem_object='CIMParameter',
                name=u'Chicken',
                type=u'string',
                reference_class=None,
                embedded_object=None,
                is_array=False,
                array_size=None,
                qualifiers=yaml_ordereddict(),
            ),
        ),
        None, None, True
    ),
    # TODO: Add testcase for reference parameter
    # TODO: Add testcase for embedded object parameter
    # TODO: Add testcase for array parameter
    (
        "CIMQualifier object",
        dict(
            obj=CIMQualifier('Chicken', type='string', value='Ham'),
            exp_yaml=dict(
                pywbem_object='CIMQualifier',
                name=u'Chicken',
                type=u'string',
                value=u'Ham',
                propagated=None,
                tosubclass=None,
                toinstance=None,
                overridable=None,
                translatable=None,
            ),
        ),
        None, None, True
    ),
    (
        "CIMQualifierDeclaration object",
        dict(
            obj=CIMQualifierDeclaration('Chicken', type='string', value='Ham'),
            exp_yaml=dict(
                pywbem_object='CIMQualifierDeclaration',
                name=u'Chicken',
                type=u'string',
                value=u'Ham',
                is_array=False,
                array_size=None,
                scopes=yaml_ordereddict(),
                tosubclass=None,
                toinstance=None,
                overridable=None,
                translatable=None,
            ),
        ),
        None, None, True
    ),
    # TODO: Add testcase for array qualifier decl
    (
        "pull_path_result_tuple object for exhausted open/pull",
        dict(
            obj=pull_path_result_tuple(
                [],
                True,
                None,
            ),
            exp_yaml=dict(
                paths=[],
                eos=True,
                context=None,
            ),
        ),
        None, None, True
    ),
    (
        "pull_path_result_tuple object for non-exhausted open/pull",
        dict(
            obj=pull_path_result_tuple(
                [],
                False,
                ('test_rtn_context', 'root/cim_namespace'),
            ),
            exp_yaml=dict(
                paths=[],
                eos=False,
                context=['test_rtn_context', 'root/cim_namespace'],
            ),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_TESTCLIENTRECORDER_TOYAML)
@simplified_test_function
def test_TestClientRecorder_toyaml(testcase, obj, exp_yaml):
    # pylint: disable=unused-argument
    """
    Test function for TestClientRecorder.toyaml()
    """

    try:
        fp = open(os.devnull, 'w')

        recorder = _TestClientRecorder(fp)
        recorder.reset()
        recorder.enable()

        # The code to be tested
        act_yaml = recorder.toyaml(obj)

    finally:
        fp.close()

    assert act_yaml == exp_yaml


TESTCASES_TESTCLIENTRECORDER_RECORD = [

    # Testcases for TestClientRecorder.toyaml()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * op_name: Name of the operation
    # * op_kwargs: Input arguments for the operation as kwargs dict
    # * op_result: Return value of the operation
    # * op_exc: Exception raised by the operation, or None
    # * exp_yaml_items: partial set of items that should be compared against
    #   the items in the first entry of the generated test YAML file.

    (
        "InvokeMethod of instance method",
        'InvokeMethod',
        dict(
            namespace='cim/blah',
            MethodName='Blah',
            ObjectName=CIMInstanceName('C1'),
            Params=[
                CIMParameter('P1', type='string', value='abc'),
            ],
        ),
        Uint32(42),
        None,
        dict(
            name='InvokeMethod',
            pywbem_request=dict(
                url='http://acme.com:80',
                creds=['username', 'password'],
                debug=False,
                timeout=10,
                namespace='root/cimv2',
                operation=dict(
                    pywbem_method='InvokeMethod',
                    namespace='cim/blah',
                    MethodName='Blah',
                    ObjectName=dict(
                        pywbem_object='CIMInstanceName',
                        classname='C1',
                        keybindings=yaml_ordereddict(),
                        namespace=None,
                        host=None,
                    ),
                    Params=[
                        dict(
                            pywbem_object='CIMParameter',
                            name='P1',
                            type='string',
                            array_size=None,
                            is_array=False,
                            reference_class=None,
                            embedded_object=None,
                            qualifiers=yaml_ordereddict(),
                        )
                    ],
                ),
            ),
            pywbem_response=dict(
                result=42,
            ),
        ),
    ),
    (
        "GetInstance on an instance",
        'GetInstance',
        dict(
            namespace='cim/blah',
            InstanceName=CIMInstanceName('C1'),
            LocalOnly=False,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'],
        ),
        CIMInstance('C1'),
        None,
        dict(
            name='GetInstance',
            pywbem_request=dict(
                url='http://acme.com:80',
                creds=['username', 'password'],
                debug=False,
                timeout=10,
                namespace='root/cimv2',
                operation=dict(
                    pywbem_method='GetInstance',
                    namespace='cim/blah',
                    IncludeClassOrigin=True,
                    IncludeQualifiers=True,
                    InstanceName=dict(
                        pywbem_object='CIMInstanceName',
                        classname='C1',
                        keybindings=yaml_ordereddict(),
                        namespace=None,
                        host=None,
                    ),
                    LocalOnly=False,
                    PropertyList=['propertyblah'],
                ),
            ),
            pywbem_response=dict(
                result=dict(
                    pywbem_object='CIMInstance',
                    classname='C1',
                    properties=yaml_ordereddict(),
                    qualifiers=yaml_ordereddict(),
                    path=None,
                ),
            ),
        ),
    ),
    (
        "CreateInstance with properties of all types",
        'CreateInstance',
        dict(
            namespace='cim/blah',
            NewInstance=CIMINSTANCE_ALL_OBJ,
        ),
        CIMInstanceName('C1'),
        None,
        dict(
            name='CreateInstance',
            pywbem_request=dict(
                url='http://acme.com:80',
                creds=['username', 'password'],
                debug=False,
                timeout=10,
                namespace='root/cimv2',
                operation=dict(
                    pywbem_method='CreateInstance',
                    namespace='cim/blah',
                    NewInstance=CIMINSTANCE_ALL_TCYAML,
                ),
            ),
            pywbem_response=dict(
                result=dict(
                    pywbem_object='CIMInstanceName',
                    classname='C1',
                    keybindings=yaml_ordereddict(),
                    namespace=None,
                    host=None,
                ),
            ),
        ),
    ),
    (
        "OpenEnumerateInstancePaths operation",
        'OpenEnumerateInstancePaths',
        dict(
            namespace='cim/blah',
            ClassName='CIM_BLAH',
            FilterQueryLanguage='WQL',
            FilterQuery='Property = 3',
            OperationTimeout=40,
            ContinueOnError=False,
            MaxObjectCount=100,
        ),
        pull_path_result_tuple(
            [],
            False,
            ('test_rtn_context', 'root/cim_namespace'),
        ),
        None,
        dict(
            name='OpenEnumerateInstancePaths',
            pywbem_request=dict(
                url='http://acme.com:80',
                creds=['username', 'password'],
                debug=False,
                timeout=10,
                namespace='root/cimv2',
                operation=dict(
                    pywbem_method='OpenEnumerateInstancePaths',
                    namespace='cim/blah',
                    ClassName='CIM_BLAH',
                    ContinueOnError=False,
                    FilterQuery='Property = 3',
                    FilterQueryLanguage='WQL',
                    MaxObjectCount=100,
                    OperationTimeout=40,
                ),
            ),
            pywbem_response=dict(
                pullresult=dict(
                    context=['test_rtn_context', 'root/cim_namespace'],
                    eos=False,
                    paths=[],
                ),
            ),
        ),
    ),

]


@pytest.mark.parametrize(
    "desc, op_name, op_kwargs, op_result, op_exc, exp_yaml_items",
    TESTCASES_TESTCLIENTRECORDER_RECORD)
def test_TestClientRecorder_record(
        desc, op_name, op_kwargs, op_result, op_exc, exp_yaml_items,
        test_client_recorder):
    # pylint: disable=redefined-outer-name
    """
    Record a single operation using the test client recorder and verify the
    generated YAML file.
    """

    # The code to be tested
    if op_name.startswith('Open') or op_name.startswith('Pull'):
        test_client_recorder.reset(pull_op=True)
    test_client_recorder.stage_pywbem_args(method=op_name, **op_kwargs)
    test_client_recorder.stage_pywbem_result(op_result, op_exc)
    test_client_recorder.record_staged()

    if DEBUG_TEST_YAML_FILE:
        print("\nDebug: Test client YAML file for testcase: {}".
              format(desc))
        if sys.platform == 'win32':
            os.system('type {}'.format(TEST_YAML_FILE))
        else:
            os.system('cat {}'.format(TEST_YAML_FILE))

    # Verify the generated test client YAML file
    test_yaml = load_recorder_yaml_file()
    assert len(test_yaml) == 1
    test_yaml_record = test_yaml[0]
    for key in exp_yaml_items:
        exp_value = exp_yaml_items[key]
        assert_yaml_equal(test_yaml_record[key], exp_value, key)

    cleanup_recorder_yaml_file()


def assert_yaml_equal(act_value, exp_value, location):
    """
    Assert that two YAML objects are equal, whereby dictionaries are compared
    such that the use of an OrderedDict on the side of the expected value
    takes ordering into account, and any other dict type does not.
    """
    if isinstance(exp_value, OrderedDict):
        # Ordered comparison.
        assert act_value == exp_value, \
            "Unexpected YAML object at {}".format(location)
    elif isinstance(exp_value, dict):
        # Ensure unordered comparison by converting to dict.
        # Note that starting with Python 3.7, dict maintains order but equality
        # comparison does not take order into account.
        act_value2 = dict(act_value)
        assert act_value2 == exp_value, \
            "Unexpected YAML object at {}".format(location)
    else:
        assert act_value == exp_value, \
            "Unexpected YAML object at {}".format(location)


################################################################
#
#           LogOperationRecorder tests
#
################################################################

class BaseLogOperationRecorderTests(object):
    """
    Test the LogOperationRecorder functions. Creates log entries and
    uses testfixture to validate results
    """

    def setup_method(self):
        """
        Setup that is run before each test method.
        """
        # Shut down any existing logger and reset WBEMConnection and
        # reset WBEMConnection class attributes
        # pylint: disable=protected-access
        WBEMConnection._reset_logging_config()
        logging.shutdown()
        # NOTE We do not clean up handlers or logger names already defined.
        #      That should not affect the tests.

    def recorder_setup(self, detail_level=None):
        """
        Setup the recorder for a defined max output size
        """

        configure_logger('api', log_dest='file',
                         detail_level=detail_level,
                         log_filename=TEST_OUTPUT_LOG, propagate=True)

        configure_logger('http', log_dest='file',
                         detail_level=detail_level,
                         log_filename=TEST_OUTPUT_LOG, propagate=True)

        # Define an attribute that is a single LogOperationRecorder to be used
        # in some of the tests.  Note that if detail_level is dict it is used
        # directly.
        if isinstance(detail_level, (six.string_types, six.integer_types)):
            detail_level = {'api': detail_level, 'http': detail_level}

        # pylint: disable=attribute-defined-outside-init
        # Set a conn id into the connection. Saves testing the connection
        # log for each test.
        self.test_recorder = LogOperationRecorder('test_id',
                                                  detail_levels=detail_level)
        # pylint: disable=protected-access
        self.test_recorder.reset()
        self.test_recorder.enable()

    def teardown_method(self):
        """
        Teardown that is run after each test method.
        """
        LogCapture.uninstall_all()
        logging.shutdown()
        # remove any existing log file
        if os.path.isfile(TEST_OUTPUT_LOG):
            os.remove(TEST_OUTPUT_LOG)


class Test_LOR_Connections(BaseLogOperationRecorderTests):
    """
    Test the LogOperationRecorder with just connections, without any operations.
    """

    def test_connection_1(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test log of a WBEMConnection object with default parameters.
        """

        # Fake the connection to create a fixed data environment
        conn = WBEMConnection('http://blah:5988')

        configure_logger('api', log_dest='file', detail_level='all',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace='root/cimv2', "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )

    def test_connection_2(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test log of a WBEMConnection object with most parameters specified,
        and detail level 'all'.
        """

        conn = WBEMConnection('http://blah:5988',
                              default_namespace='root/blah',
                              creds=('username', 'password'),
                              no_verification=True,
                              timeout=10,
                              use_pull_operations=True,
                              stats_enabled=True)

        configure_logger('api', log_dest='file', detail_level='all',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah:5988', "
            "creds=('username', ...), "
            "conn_id={0!A}, "
            "default_namespace='root/blah', "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=True, "
            "timeout=10, "
            "use_pull_operations=True, "
            "stats_enabled=True, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )

    def test_connection_summary(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test log of a WBEMConnection object with most parameters specified,
        and detail level 'summary'.
        """

        conn = WBEMConnection('http://blah:5988',
                              default_namespace='root/blah',
                              creds=('username', 'password'),
                              no_verification=True,
                              timeout=10,
                              use_pull_operations=True,
                              stats_enabled=True)

        configure_logger('api', log_dest='file', detail_level='summary',
                         connection=conn, log_filename=TEST_OUTPUT_LOG,
                         propagate=True)

        conn_id = conn.conn_id
        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah:5988', "
            "creds=('username', ...), "
            "default_namespace='root/blah', "
            "...)",
            conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
        )


class Test_LOR_PywbemResults(BaseLogOperationRecorderTests):
    """
    Test the LogOperationRecorder by staging pywbem results of operations.
    """

    def test_result_exception(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log None return, HTTPError exception.
        """

        self.recorder_setup(detail_level=10)

        # Note: cimerror is the CIMError HTTP header field
        exc = HTTPError(500, "Fake Reason", cimerror="Fake CIMError")

        self.test_recorder.stage_pywbem_result(None, exc)

        result_exc = "Exception:test_id None('HTTPError...)"

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    def test_result_exception_all(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log None return, HTTPError exception.
        """

        self.recorder_setup(detail_level='all')

        # Note: cimerror is the CIMError HTTP header field
        exc = HTTPError(500, "Fake Reason", cimerror="Fake CIMError")

        self.test_recorder.stage_pywbem_result(None, exc)

        result_exc = _format(
            "Exception:test_id None('HTTPError({0})')",
            exc)

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    def test_result_getinstance(self, capture, instancename_tuple):
        # pylint: disable=redefined-outer-name
        """
        Emulates call to getInstance to test parameter processing.
        Currently creates the pywbem_request component.
        """

        instancename = instancename_tuple[0]
        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=instancename,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        result_req = \
            "Request:test_id GetInstance(" \
            "IncludeClassOrigin=True, " \
            "IncludeQualifiers=True, " \
            "InstanceName={ip!r}, " \
            "LocalOnly=True, " \
            "PropertyList=['propertyblah'])". \
            format(ip=instancename)

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
        )

    @pytest.mark.parametrize(
        "detail_level", [10, 1000, 'all'])
    def test_result_instance(self, capture, instance_tuple, detail_level):
        # pylint: disable=redefined-outer-name
        """
        Test the staging of a CIM instance with different detail_level values.
        """

        instance = instance_tuple[0]
        self.recorder_setup(detail_level=detail_level)
        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        exp_str = repr(instance)
        if detail_level == 'all':
            # use it in its full length
            pass
        elif len(exp_str) <= detail_level:
            # it fits the desired detail
            pass
        else:
            # shorten it to the desired detail
            exp_str = "{}...".format(exp_str[:detail_level])

        result_ret = "Return:test_id None({})".format(exp_str)

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_result_instance_paths(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the staging of a CIM instance with path with detail_level 'paths'.
        """

        instance_wp = instance_wp_tuple[0]
        self.recorder_setup(detail_level='paths')
        exc = None

        self.test_recorder.stage_pywbem_result(instance_wp, exc)

        path_str = "'{}'".format(instance_wp.path.to_wbem_uri())
        result_ret = "Return:test_id None({})".format(path_str)

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_result_instances_paths(self, capture, instances_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the staging of multiple CIM instances with path with
        detail_level 'paths'.
        """

        instances_wp = instances_wp_tuple[0]
        self.recorder_setup(detail_level='paths')
        exc = None

        self.test_recorder.stage_pywbem_result(instances_wp, exc)

        path_strs = ["'{}'".format(inst.path.to_wbem_uri())
                     for inst in instances_wp]
        result_ret = "Return:test_id None({})".format(', '.join(path_strs))

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_result_pull_instances_paths(self, capture, instances_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the staging of multiple CIM instances with path with
        detail_level 'paths'.
        """

        instances_wp = instances_wp_tuple[0]
        self.recorder_setup(detail_level='paths')
        exc = None

        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_inst_result_tuple(
            instances_wp, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        path_strs = ["'{}'".format(inst.path.to_wbem_uri())
                     for inst in instances_wp]
        result_ret = (
            "Return:test_id None(pull_inst_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "instances={}))".format(', '.join(path_strs))
        )

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )


class Test_LOR_HTTPRequests(BaseLogOperationRecorderTests):
    """
    Test the LogOperationRecorder by staging HTTP requests of operations.
    """

    def build_http_request(self, instancename):
        """
        Build the HTTP request for a GetInstance operation for the specified
        instance name, for use by the following tests.
        """
        headers = OrderedDict([
            ('CIMOperation', 'MethodCall'),
            ('CIMMethod', 'GetInstance'),
            ('CIMObject', 'root/cimv2')])
        url = 'http://blah:5988'
        method = 'POST'
        target = '/cimom'

        payload = (
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            u'<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            u'<SIMPLEREQ>\n'
            u'<IMETHODCALL NAME="GetInstance">\n'
            u'<LOCALNAMESPACEPATH>\n'
            u'<NAMESPACE NAME="root"/>\n'
            u'<NAMESPACE NAME="cimv2"/>\n'
            u'</LOCALNAMESPACEPATH>\n'
            u'<IPARAMVALUE NAME="InstanceName">\n'
            u'{ip}\n'
            u'</IPARAMVALUE>\n'
            u'<IPARAMVALUE NAME="LocalOnly">\n'
            u'<VALUE>FALSE</VALUE>\n'
            u'</IPARAMVALUE>\n'
            u'</IMETHODCALL>\n'
            u'</SIMPLEREQ>\n'
            u'</MESSAGE>\n'
            u'</CIM>)'.
            format(ip=instancename.tocimxmlstr(indent=0))
        )

        return url, target, method, headers, payload

    def test_stage_http_request_all(self, capture, instancename_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test stage of http_request log with detail_level='all'
        """

        instancename = instancename_tuple[0]
        self.recorder_setup(detail_level='all')
        url, target, method, headers, payload = \
            self.build_http_request(instancename)

        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah:5988 "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2' "
            "{}".
            format(logged_payload(payload)))

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )

    def test_stage_http_request_summary(self, capture, instancename_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test http request log record with summary as detail level
        """

        instancename = instancename_tuple[0]
        self.recorder_setup(detail_level='summary')
        url, target, method, headers, payload = \
            self.build_http_request(instancename)
        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah:5988 "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2' "
            "''")

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )

    def test_stage_http_request_int(self, capture, instancename_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test http log record with integer as detail_level
        """

        instancename = instancename_tuple[0]
        self.recorder_setup(detail_level=10)
        url, target, method, headers, payload = \
            self.build_http_request(instancename)

        self.test_recorder.stage_http_request('test_id', 11, url, target,
                                              method, headers, payload)

        result_req = (
            "Request:test_id POST /cimom 11 http://blah:5988 "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetInstance' "
            "CIMObject:'root/cimv2' "
            "'<?xml vers...'")

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_req),
        )


class Test_LOR_HTTPResponses(BaseLogOperationRecorderTests):
    """
    Test the LogOperationRecorder by staging HTTP responses of operations.
    """

    def build_http_response(self, instance):
        """
        Build the HTTP response for a GetInstance operation for the specified
        instance, for use by the following tests.
        Also, perform part 1 of the response staging.
        """
        body = (
            u'<?xml version="1.0" encoding="utf-8" ?>\n'
            u'<CIM CIMVERSION="2.0" DTDVERSION="2.0">\n'
            u'<MESSAGE ID="1001" PROTOCOLVERSION="1.0">\n'
            u'<SIMPLERSP>\n'
            u'<IMETHODRESPONSE NAME="GetInstance">\n'
            u'<IRETURNVALUE>\n'
            u'{i}\n'
            u'</IRETURNVALUE>\n'
            u'</IMETHODRESPONSE>\n'
            u'</SIMPLERSP>\n'
            u'</MESSAGE>\n'
            u'</CIM>)\n'.
            format(i=instance.tocimxmlstr(indent=0))
        )
        headers = OrderedDict([
            ('Content-type', 'application/xml; charset="utf-8"'),
            ('Content-length', str(len(body)))])
        status = 200
        reason = ""
        version = 11
        self.test_recorder.stage_http_response1('test_id', version,
                                                status, reason, headers)
        return body

    def test_stage_http_response_all(self, capture, instance_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test http response log record with 'all' detail_level
        """

        instance = instance_tuple[0]
        self.recorder_setup(detail_level='all')

        body = self.build_http_response(instance)

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'{bl}' "
            "{b}".
            format(bl=len(body), b=logged_payload(body))
        )

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )

    def test_stage_http_response_summary(self, capture, instance_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test http response log record with 'all' detail_level
        """

        instance = instance_tuple[0]
        self.recorder_setup(detail_level='summary')

        body = self.build_http_response(instance)

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'{bl}' "
            "''".
            format(bl=len(body))
        )

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )

    def test_stage_http_response_int(self, capture, instance_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test http response log record with 'all' detail_level
        """

        instance = instance_tuple[0]
        self.recorder_setup(detail_level=30)

        body = self.build_http_response(instance)

        self.test_recorder.stage_http_response2(body)

        result_resp = (
            "Response:test_id 200: 11 "
            "Content-type:'application/xml; charset=\"utf-8\"' "
            "Content-length:'{bl}' "
            '\'<?xml version="1.0" encoding="...\''.
            format(bl=len(body))
        )

        capture.check(
            ('pywbem.http.test_id', 'DEBUG', result_resp),
        )


class Test_LOR_PywbemArgsResults(BaseLogOperationRecorderTests):
    """
    Test the LogOperationRecorder by staging pywbem args and results of
    operations.
    """

    def test_getinstance(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for get instance
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=instancename,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        exc = None
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance(IncludeCla...)")

        result_ret = (
            "Return:test_id GetInstance(CIMInstanc...)")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_getinstance_exception(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for get instance
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        self.recorder_setup(detail_level=11)

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=instancename,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        instance = None
        exc = CIMError(6, "Fake CIMError")
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance(IncludeClas...)")

        result_exc = (
            "Exception:test_id GetInstance('CIMError(6...)")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    def test_getinstance_exception_all(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for get instance CIMError exception
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=instancename)

        instance = None
        exc = CIMError(6, "Fake CIMError")
        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance("
            "InstanceName={!r})".
            format(instancename))

        result_exc = _format(
            "Exception:test_id GetInstance('CIMError({0})')", exc)

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    def test_getinstance_result_all(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for get instance
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='GetInstance',
            InstanceName=instancename,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        exc = None

        self.test_recorder.stage_pywbem_result(instance, exc)

        result_req = (
            "Request:test_id GetInstance("
            "IncludeClassOrigin=True, "
            "IncludeQualifiers=True, "
            "InstanceName={!r}, "
            "LocalOnly=True, "
            "PropertyList=['propertyblah'])".
            format(instancename))

        result_ret = (
            "Return:test_id GetInstance({!r})".
            format(instance))

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_enuminstances_result(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for enumerate instances
        """

        instance = instance_wp_tuple[0]
        classname = instance.classname

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstances',
            ClassName=classname,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        exc = None

        self.test_recorder.stage_pywbem_result([instance, instance], exc)

        result_req = (
            "Request:test_id EnumerateInstances(ClassName=...)")

        result_ret = (
            "Return:test_id EnumerateInstances([CIMInstan...)")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_enuminstancenames_result(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for enumerate instances
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path
        classname = instance.classname

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstanceNames',
            ClassName=classname,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'blah2'])

        exc = None

        self.test_recorder.stage_pywbem_result(
            [instancename, instancename], exc)

        result_req = (
            "Request:test_id EnumerateInstanceNames(ClassName=...)")

        result_ret = (
            "Return:test_id EnumerateInstanceNames([CIMInstan...)")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_openenuminstances_result_all(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for enumerate instances. Returns no instances.
        """

        classname = 'CIM_Foo'

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='OpenEnumerateInstances',
            ClassName=classname,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah'])

        exc = None

        result = []
        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_inst_result_tuple(result, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        result_req = (
            "Request:test_id OpenEnumerateInstances("
            "ClassName={!r}, "
            "IncludeClassOrigin=True, "
            "IncludeQualifiers=True, "
            "LocalOnly=True, "
            "PropertyList=['propertyblah'])".
            format(classname))

        result_ret = (
            "Return:test_id OpenEnumerateInstances("
            "pull_inst_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "instances=[]))")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_openenuminstances_all(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for enumerate instances paths with
        data in the paths component
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path
        classname = instance.classname

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level='all')

        self.test_recorder.stage_pywbem_args(
            method='OpenEnumerateInstancePaths',
            ClassName=classname,
            FilterQueryLanguage='FQL',
            FilterQuery='SELECT A from B',
            OperationTimeout=10,
            ContinueOnError=None,
            MaxObjectCount=100)

        result = [instancename, instancename]
        exc = None

        context = ('test_rtn_context', 'root/blah')
        result_tuple = pull_path_result_tuple(result, False, context)

        self.test_recorder.stage_pywbem_result(result_tuple, exc)

        result_req = (
            "Request:test_id OpenEnumerateInstancePaths("
            "ClassName={!r}, "
            "ContinueOnError=None, "
            "FilterQuery='SELECT A from B', "
            "FilterQueryLanguage='FQL', "
            "MaxObjectCount=100, "
            "OperationTimeout=10)".
            format(classname))

        result_ret = (
            "Return:test_id OpenEnumerateInstancePaths("
            "pull_path_result_tuple("
            "context=('test_rtn_context', 'root/blah'), eos=False, "
            "paths={!r}))".
            format(result))

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_associators_result(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for Associators that returns nothing
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=10)

        self.test_recorder.stage_pywbem_args(
            method='Associators',
            InstanceName=instancename,
            AssocClass='BLAH_Assoc',
            ResultClass='BLAH_Result',
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'propertyblah2'])
        exc = None

        self.test_recorder.stage_pywbem_result([], exc)

        result_req = (
            "Request:test_id Associators(AssocClass...)")

        result_ret = (
            "Return:test_id Associators([])")

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_associators_result_exception(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test the ops result log for associators that returns exception
        """

        # set recorder to limit response to length of 10
        self.recorder_setup(detail_level=11)

        exc = CIMError(6, "Fake CIMError")

        self.test_recorder.stage_pywbem_result([], exc)

        result_exc = "Exception:test_id None('CIMError(6...)"

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_exc),
        )

    def test_invokemethod_int(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test invoke method log
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        self.recorder_setup(detail_level=11)

        return_val = 0
        params = [('StringParam', 'Spotty'),
                  ('Uint8', Uint8(1)),
                  ('Sint8', Sint8(2))]

        self.test_recorder.stage_pywbem_args(method='InvokeMethod',
                                             MethodName='Blah',
                                             ObjectName=instancename,
                                             Params=OrderedDict(params))

        self.test_recorder.stage_pywbem_result((return_val, params),
                                               None)

        result_req = "Request:test_id InvokeMethod(MethodName=...)"

        result_ret = "Return:test_id InvokeMethod((0, [('Stri...)"

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    def test_invokemethod_summary(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Test invoke method log
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path

        self.recorder_setup(detail_level='summary')

        return_val = 0
        params = [('StringParam', 'Spotty'),
                  ('Uint8', Uint8(1)),
                  ('Sint8', Sint8(2))]

        self.test_recorder.stage_pywbem_args(method='InvokeMethod',
                                             MethodName='Blah',
                                             ObjectName=instancename,
                                             Params=OrderedDict(params))

        self.test_recorder.stage_pywbem_result((return_val, params),
                                               None)

        result_req = (
            "Request:test_id InvokeMethod("
            "MethodName='Blah', "
            "ObjectName={!r}, "
            "Params=OrderedDict(["
            "('StringParam', 'Spotty'), "
            "('Uint8', Uint8(cimtype='uint8', minvalue=0, maxvalue=255, 1)), "
            "('Sint8', Sint8(cimtype='sint8', minvalue=-128, maxvalue=127, 2))"
            "]))".
            format(instancename))

        result_ret = "Return:test_id InvokeMethod(tuple )"

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    # TODO add tests for all for invoke method.


class TestExternLoggerDef(BaseLogOperationRecorderTests):
    """
    Test configuring loggers above level of our loggers
    """

    def test_root_logger(self, capture, instance_wp_tuple):
        # pylint: disable=redefined-outer-name
        """
        Create a logger using logging.basicConfig and generate logs
        """

        instance = instance_wp_tuple[0]
        instancename = instance.path
        classname = instance.classname

        logging.basicConfig(filename=TEST_OUTPUT_LOG, level=logging.DEBUG)

        detail_level_int = 10
        detail_level = {'api': detail_level_int, 'http': detail_level_int}

        # pylint: disable=attribute-defined-outside-init
        self.test_recorder = LogOperationRecorder('test_id',
                                                  detail_levels=detail_level)

        # We don't activate because we are not using wbemconnection, just
        # the recorder calls.
        self.test_recorder.stage_pywbem_args(
            method='EnumerateInstanceNames',
            ClassName=classname,
            LocalOnly=True,
            IncludeQualifiers=True,
            IncludeClassOrigin=True,
            PropertyList=['propertyblah', 'blah2'])

        exc = None
        result = [instancename, instancename]

        self.test_recorder.stage_pywbem_result(result, exc)

        result_req = "Request:test_id EnumerateInstanceNames(ClassName=...)"

        result_ret = "Return:test_id EnumerateInstanceNames([CIMInstan...)"

        capture.check(
            ('pywbem.api.test_id', 'DEBUG', result_req),
            ('pywbem.api.test_id', 'DEBUG', result_ret),
        )

    @pytest.mark.skip("Test unreliable exception not always the same")
    def test_pywbem_logger(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test executing a connection with an externally created handler. Note
        that this test inconsistently reports the text of the exception
        in that sometimes adds the message  "[Errno 113] No route to host" and
        sometimes the message "timed out".
        """

        logger = logging.getLogger('pywbem')
        handler = logging.FileHandler(TEST_OUTPUT_LOG)
        handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        configure_logger('api', detail_level='summary', connection=True,
                         propagate=True)
        try:
            conn = WBEMConnection('http://blah:5988', timeout=1)

            conn.GetClass('blah')
        except Exception:  # pylint: disable=broad-except
            pass

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)
        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_exc = _format(
            "Exception:{0} GetClass('ConnectionError(Socket error: "
            "[Errno 113] No route to host)')",
            conn_id)

        result_con = _format(
            "Connection:{0} WBEMConnection("
            "url='http://blah:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace='root/cimv2', "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=1, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder'])",
            conn_id)

        result_req = _format(
            "Request:{0} GetClass("
            "ClassName='blah', "
            "IncludeClassOrigin=None, "
            "IncludeQualifiers=None, "
            "LocalOnly=None, "
            "PropertyList=None, "
            "namespace=None)",
            conn_id)

        result_hreq = _format(
            "Request:{0} POST /cimom 11 http://blah:5988 "
            "CIMOperation:'MethodCall' "
            "CIMMethod:'GetClass' "
            "CIMObject:u'root/cimv2'\n"
            '<?xml version="1.0" encoding="utf-8" ?>\n'
            '<CIM CIMVERSION="2.0" DTDVERSION="2.0">'
            '<MESSAGE ID="1001" PROTOCOLVERSION="1.0">'
            '<SIMPLEREQ>'
            '<IMETHODCALL NAME="GetClass">'
            '<LOCALNAMESPACEPATH>'
            '<NAMESPACE NAME="root"/>'
            '<NAMESPACE NAME="cimv2"/>'
            '</LOCALNAMESPACEPATH>'
            '<IPARAMVALUE NAME="ClassName">'
            '<CLASSNAME NAME="blah"/>'
            '</IPARAMVALUE>'
            '</IMETHODCALL>'
            '</SIMPLEREQ>'
            '</MESSAGE>'
            '</CIM>',
            conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (http_exp_log_id, 'DEBUG', result_hreq),
            (api_exp_log_id, 'DEBUG', result_exc),
        )


class TestLoggingEndToEnd(BaseLogOperationRecorderTests):
    """
    Test the examples documented in _logging.py document string. Because
    there is not logging until the first http request, these extend the
    examples to actually issue a request.  Therefore, they use the mocker
    to emulate a server

    """

    def build_repo(self, namespace):
        """Build a CIM repository using FakedWBEMConnection"""
        skip_if_moftab_regenerated()

        schema = install_test_dmtf_schema()
        partial_schema = """
            #pragma locale ("en_US")
            #pragma include ("Interop/CIM_ObjectManager.mof")
            """

        conn = FakedWBEMConnection(default_namespace=namespace)
        conn.compile_mof_string(partial_schema, namespace=namespace,
                                search_paths=[schema.schema_mof_dir])
        return conn

    def test_1(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure the "pywbem.api" logger for summary information output to a
        file and activate that logger for all subsequently created
        :class:`~pywbem.WBEMConnection` objects.
        NOTE: We changed from example to log to file and use log_captyre
        """
        skip_if_moftab_regenerated()

        # setup schema in test because we configure before we create the
        # connection in this test.
        namespace = 'root/blah'
        schema = install_test_dmtf_schema()
        partial_schema = """
            #pragma locale ("en_US")
            #pragma include ("Interop/CIM_ObjectManager.mof")
            """
        conn = FakedWBEMConnection(default_namespace=namespace)
        conn.compile_mof_string(partial_schema, namespace=namespace,
                                search_paths=[schema.schema_mof_dir])

        configure_logger('api', log_dest='file',
                         detail_level='summary',
                         log_filename=TEST_OUTPUT_LOG,
                         connection=conn, propagate=True)
        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "default_namespace={1!A}, ...))",
            conn_id, namespace)

        result_req = _format(
            "Request:{0} GetClass("
            "ClassName='CIM_ObjectManager', "
            "IncludeClassOrigin=None, "
            "IncludeQualifiers=None, "
            "LocalOnly=None, "
            "PropertyList=None, "
            "namespace={1!A})",
            conn_id, namespace)

        result_ret = _format(
            "Return:{0} GetClass(CIMClass CIM_ObjectManager)",
            conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret)
        )

    def test_2(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure and activate a single :class:`~pywbem.WBEMConnection` object
        logger for output of summary information for both "pywbem.api" and
        "pywbem.http"::
        Differs from example in that we set detail_level to limit output for
        test
        """
        namespace = 'root/blah'
        conn = self.build_repo(namespace)
        configure_logger('all', log_dest='file',
                         log_filename=TEST_OUTPUT_LOG,
                         detail_level=10,
                         connection=conn, propagate=True)
        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    def test_3(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test.
        Does not produce http request/response info because when using the
        pywbem_mock, no http is generated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('all', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    def test_4(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test
        """
        logging.basicConfig(filename='basicconfig.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('api', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret),
        )

    def test_5(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a single pywbem connection with standard Python logger
        methods by defining the root logger with basicConfig
        Differs from example in that we set detail_level to limit output for
        test. The basic logger should have no effect since pywbem sets the
        NULLHandler at the pywbem level.
        """
        logging.basicConfig(filename='basicconfig.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('api', detail_level=10, connection=conn,
                         propagate=True)

        # logging_tree_printout()

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        api_exp_log_id = 'pywbem.api.{0}'.format(conn_id)

        result_con = _format("Connection:{0} FakedWBEMC...", conn_id)

        result_req = _format("Request:{0} GetClass(ClassName=...)", conn_id)

        result_ret = _format("Return:{0} GetClass(CIMClass(c...)", conn_id)

        capture.check(
            (api_exp_log_id, 'DEBUG', result_con),
            (api_exp_log_id, 'DEBUG', result_req),
            (api_exp_log_id, 'DEBUG', result_ret)
        )

    def test_6(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest='file',
                         log_filename=TEST_OUTPUT_LOG,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "disable_pull_operations=False "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        capture.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    def test_7(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a http logger with detail_level='all', log_dest=none
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        logging.basicConfig(filename='example.log', level=logging.DEBUG)
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "disable_pull_operations=False "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        capture.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    def test_8(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """
        msg_format = '%(asctime)s-%(name)s-%(message)s'
        handler = logging.FileHandler(TEST_OUTPUT_LOG, encoding="UTF-8")
        handler.setFormatter(logging.Formatter(msg_format))
        logger = logging.getLogger('pywbem')
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "disable_pull_operations=False "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        capture.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    def test_9(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """

        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all', log_dest=None,
                         connection=conn, propagate=True)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "disable_pull_operations=False "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        capture.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    def test_10(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Configure a http logger with detail_level='all' and
        a log config with just 'http'  (which produces no output because the
        mocker does not generate http requests or responses) and at the
        same time setting the logging basic logger which we do not use.
        This test ability to generate connection log when only http log is
        activated.
        """

        namespace = 'root/blah'
        # Define the detail_level and WBEMConnection object to activate.
        configure_logger('http', detail_level='all',
                         connection=True, propagate=True)
        conn = self.build_repo(namespace)

        conn.GetClass('CIM_ObjectManager', namespace=namespace)

        conn_id = conn.conn_id

        http_exp_log_id = 'pywbem.http.{0}'.format(conn_id)

        result_con = _format(
            "Connection:{0} FakedWBEMConnection("
            "response_delay=None, "
            "disable_pull_operations=None "
            "super=WBEMConnection("
            "url='http://FakedUrl:5988', "
            "creds=None, "
            "conn_id={0!A}, "
            "default_namespace={1!A}, "
            "x509=None, "
            "ca_certs=None, "
            "no_verification=False, "
            "timeout=None, "
            "use_pull_operations=False, "
            "stats_enabled=False, "
            "recorders=['LogOperationRecorder']))",
            conn_id, namespace)

        capture.check(
            (http_exp_log_id, 'DEBUG', result_con)
        )

    def test_err(self, capture):
        # pylint: disable=redefined-outer-name
        """
        Test configure_logger exception
        """
        namespace = 'root/blah'
        conn = self.build_repo(namespace)

        # Define the detail_level and WBEMConnection object to activate.

        with pytest.raises(ValueError):
            configure_logger('api', detail_level='blah', connection=conn,
                             propagate=True)

        capture.check()



"""
End2end tests for WBEMServer class.
"""

from __future__ import absolute_import, print_function

import re

from .utils.utils import server_func_asserted, server_prop_asserted

# Note: The wbem_connection fixture uses the server_definition fixture, and
# due to the way py.test searches for fixtures, it also need to be imported.
# pylint: disable=unused-import
from .utils.pytest_extensions import wbem_connection  # noqa: F401
from .utils.pytest_extensions import server_definition  # noqa: F401
from .utils.pytest_extensions import default_namespace  # noqa: F401
# pylint: enable=unused-import

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils import import_installed
pywbem = import_installed('pywbem')
from pywbem import WBEMServer  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


def test_namespace_consistency(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the Interop namespace and all namespaces and namespace paths can
    be determined, and verify consistency between them.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    interop_ns_lower = server_prop_asserted(server, 'interop_ns').lower()
    namespaces_lower = [ns.lower()
                        for ns in server_prop_asserted(server, 'namespaces')]
    namespace_paths = server_prop_asserted(server, 'namespace_paths')

    assert interop_ns_lower in namespaces_lower, msg

    # Note: One server has been found that supports the Interop namespace, but
    # does not represent it as a CIM instance. In that case, server.namespaces
    # contains the Interop namespace while server.namespace_paths does not.
    # The following checks accomodate that.
    for ns_path in namespace_paths:
        assert ns_path.namespace.lower() == interop_ns_lower, msg
        ns_path_name_lower = ns_path['Name'].lower()
        assert ns_path_name_lower in namespaces_lower, msg


def test_namespace_getinstance(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that GetInstance on the instance paths of all namespaces succeeds.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    namespace_paths = server_prop_asserted(server, 'namespace_paths')

    for path in namespace_paths:

        inst = wbem_connection.GetInstance(
            path,
            asserted=True)

        inst_name_lower = inst['Name'].lower()
        path_name_lower = path['Name'].lower()
        assert inst_name_lower == path_name_lower, msg


def test_brand_version(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the server brand and version can be determined.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    brand = server_prop_asserted(server, 'brand')
    assert brand is not None, msg

    # Note: The version is None if it cannot be determined.
    # We check that it begins with a digit and does not contain blanks.
    version = server_prop_asserted(server, 'version')
    if version is not None:
        assert re.match(r'^[0-9][^ ]*$', version), \
            "version={0!r}; {1}".format(version, msg)


def test_cimom_inst(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the instance representing the server can be determined and that
    GetInstance on it succeeds.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    cimom_inst = server_prop_asserted(server, 'cimom_inst')

    interop_ns_lower = server_prop_asserted(server, 'interop_ns').lower()

    assert cimom_inst.path is not None, msg
    assert cimom_inst.path.namespace.lower() == interop_ns_lower, msg

    get_inst = wbem_connection.GetInstance(
        cimom_inst.path,
        asserted=True)

    assert cimom_inst == get_inst, msg


def test_profiles(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the registered profiles advertised by the server can be
    determined and that GetInstance on them succeeds.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    profile_insts = server_prop_asserted(server, 'profiles')

    interop_ns_lower = server_prop_asserted(server, 'interop_ns').lower()

    for profile_inst in profile_insts:
        assert profile_inst.path is not None, msg
        assert profile_inst.path.namespace.lower() == interop_ns_lower, msg
        assert profile_inst['RegisteredOrganization'] is not None, msg
        assert profile_inst['RegisteredName'] is not None, msg
        assert profile_inst['RegisteredVersion'] is not None, msg

        get_inst = wbem_connection.GetInstance(
            profile_inst.path,
            asserted=True)

        assert profile_inst == get_inst, msg


def test_get_selected_profiles_no_filter(
        default_namespace, wbem_connection):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test that the get_selected_profiles() method without filtering returns
    the same profiles as the profiles attribute.
    """
    wbem_connection.default_namespace = default_namespace
    server = WBEMServer(wbem_connection)
    msg = "for server at URL {0!r}".format(wbem_connection.url)

    profile_insts = server_prop_asserted(server, 'profiles')
    all_profile_insts = server_func_asserted(server, 'get_selected_profiles')

    assert len(profile_insts) == len(all_profile_insts), msg
    for inst in profile_insts:
        assert inst in all_profile_insts, msg



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
#
"""
A CIM schema is the collection of CIM qualifier declarations and CIM
classes and available  a single tested, coherent package
with a defined major, minor, update version number. The DMTF regularly
releases updates to the DMTF CIM schema that represent all of the classes
approved by the DMTF. The MOF representation of
a DMTF schema release is packaged as a single zip file and is available on the
DMTF web site ( https://www.dmtf.org/standards/cim ).

Because CIM schemas are the source of most of the classes that a WBEM server
normally implements it was important to create a simple mechanism to include
the DMTF CIM classes and qualifier declarations from a CIM schema in the
CIM repository of :class:`~pywbem_mock.FakedWBEMConnection`.

The :class:`~pywbem_mock.DMTFCIMSchema` class downloads the DMTF CIM schema
zip file defined by version from the DMTF web site into a defined directory,
unzips the MOF files into a subdirectory and makes this subdirectory
available as a property of the :class:`~pywbem_mock.DMTFCIMSchema` object.

The implementation of this class depends on all of the qualiifiers and
classes in the CIM schema having unique class names; this is always true for
DMTF CIM schema releases to assure that they can be installed in the same
namespace in a WBEM server repository.

Once aCIM schema is extracted into the individual MOF files it consumes
considerable space since it expands to almost 3000 files.  Therefore it is
logical to remove all of the individual MOF files when not being used and also
if the schema information is to be committed to be used in testing in an
environment like github. Therefore, a method
:meth:`~pywbem_mock.DMTFCIMSchema.clean` removes all of the MOF files from
local storage.  They are restored the next time the
:class:`~pywbem_mock.DMTFCIMSchema` object for the same version is constructed.

The DMTF maintains two versions of each released schema:

* Final - This schema does not contain any of the schema classes that are
  marked experimental.  It is considered the stable release.

* Experimental - Contains the classes in the final schema plus other
  classes that are considered experimental. It is considered less stable and
  the experimental classes with the Experimental qualifier are subject to
  incompatible changes .

:class:`~pywbem_mock.DMTFCIMSchema` will install either the final or
experimental schema depending on the value of the `use_experimental`
parameter.

Because it is usually necessary to compile only a subset of the DMTF CIM
classes into a CIM repository for any test, the
:class:`~pywbem_mock.DMTFCIMSchema` class provides a
:meth:`~pywbem_mock.FakedWBEMConnection.build_schema_mof` method to create a
list of MOF pragmans that includes only a subset of the classes in the
downloaded CIM Schema. The task of listing classes to be compiled is futher
simplified because the pywbem MOF compiler searches the DMTF schema for
prerequisite classes (superclasses, reference classes, and EmbeddedInstance
classes) so that generally only the leaf CIM classes required for the
repository need to be in the class list. This speeds up the process of
compiling DMTF CIM classes for any test significantly.
"""

import os
from zipfile import ZipFile
import shutil
import re
import six
try:
    from urllib.request import urlopen
except ImportError:  # py2
    from urllib2 import urlopen

from pywbem._utils import _format


__all__ = ['DMTFCIMSchema']


def build_schema_mof(class_names, schema_pragma_file):
    """
    Build a string that includes the ``#include pragma`` statements for the
    schema CIM classes defined in `class_names` using the CIM schema defined by
    this object.

    The class names in `class_names` can be just leaf classes. The pywbem
    MOF compiler will search for dependent classes including superclasses,
    and other classes referenced as the defined classes are compiled.

    It builds a compilable MOF string in the form::

        #pragma locale ("en_US")
        #pragma include ("System/CIM_ComputerSystem.mof")
        ...

    with a ``#pragma include`` for each classname in the `class_names`
    list.

    The resulting set of #pragma statements is in the same order as the
    original pragmas.  In some cases that could be important because the
    DMTF list is ordered to missing dependencies and conflicts between
    class mof compiles.

    Parameters:

      class_names (:term:`py:list` of :term:`string` or :term:`string`):
        These must be class names of classes in the DMTF CIM schema represented
        by the instance of :class:`~pywbem_mock.DMTFCIMSchema` object. This
        parameter can be a string if there is only a single class name to be
        included.

        If the returned string is compiled, the MOF compiler will search the
        directory defined by the path component of `schema_pragma_file` for
        classes referenced by each for class in this list including super
        classes, classed defined by reference properties, and classes defined
        by an embedded-object.

    schema_pragma_file (:term:`string`):
        File path defining a CIM schema pragma file for the set of
        CIM classes that make up a schema such as the DMTF schema.
        This file must contain a pragma statement for each of the
        classes defined in the schema.

    Returns:
        :term:`string`: Valid MOF containing pragma statements from the
        `CIM schema pragma file` for all of the classes in `schema_classes`.

    Raises:
        ValueError: If any of the classnames in `schema_classes` are not in
        the `CIM schema pragma file`.
    """
    if isinstance(class_names, six.string_types):
        class_names = [class_names]

    schema_lines = []
    with open(schema_pragma_file, 'r') as f:
        schema_lines = f.readlines()

    # Build list  classname/line number pairs
    classname_pattern = ' *#.*include.*/(.*)\\.mof'
    class_names_lc = [cln.lower() for cln in class_names]
    lines_list = []
    found_classes = []
    for line_no, line in enumerate(schema_lines, 0):
        result = re.search(classname_pattern, line)
        if result:
            cln = result.group(1)
            if cln.lower() in class_names_lc:
                lines_list.append(line_no)
                found_classes.append(cln)

    # Create list of the line numbers in schema_lines containing
    # pragmas that will be part of the build.
    classes_not_found = set(class_names) - set(found_classes)
    if classes_not_found:
        raise ValueError(
            _format("Classes {0!A} not in the CIM schema pragma file {1!A}",
                    ', '.join(classes_not_found), schema_pragma_file))

    # Sort the list so the result is in the same order as the pragmas
    # in the cim_schema pragma file.
    lines_list.sort()

    # Create list with line from pragma file for each entry in lines_list
    output_lines = [schema_lines[line_number] for line_number in lines_list]

    output_lines.insert(0, '#pragma locale ("en_US")\n')

    return ''.join(output_lines)


class DMTFCIMSchema(object):
    """
    :class:`DMTFCIMSchema` represents a DMTF CIM Schema downloaded from the
    DMTF web site.

    This class manages the download of the DMTF schema zip file and extraction
    of MOF files of DMTF CIM schema releases into a directory that can be used
    by :meth:`~pywbem_mock.FakedWBEMConnection.compile_schema_classes` to
    compile into a CIM repository with class and qualifier declarations from the
    schema.
    """
    def __init__(self, schema_version, schema_root_dir, use_experimental=False,
                 verbose=False):
        """
        The init method downloads the schema if the `schema_root_dir` or the
        schema zip file do not exist into the directory defined by the
        `schema_root_dir` parameter and extracts the MOF files into
        `schema_mof_dir`.

        If the schema zip file is already downloaded it simply sets the
        :class:`DMTFCIMSchema` property values.

        If the schema zip file was previously downloaded but `schema_mof_dir`
        does not exist the directory is created and the MOF files are extracted
        into this directory.

        Parameters:

          schema_version  (tuple of 3 integers (m, n, u):
            Represents the DMTF CIM schema version m.n.u where:

            * m is the DMTF CIM schema major version
            * n is the DMTF CIM schema minor version
            * u is the DMTF CIM schema update version

            This must represent a DMTF CIM schema that is available from the
            DMTF web site.

          schema_root_dir (:term:`string`):
            Path name of the schema root directory into which the DMTF CIM
            schema zip file is downloaded and in which a subdirectory for the
            extracted MOF files for the schema version defined is created

            Multiple DMTF CIM schemas may be maintained in the same
            `schema_root_dir` simultaneously because the MOF for each schema is
            extracted into a subdirectory identified by the schema version
            information See :attr:`schema_mof_dir`.

          use_experimental (:class:`py:bool`):
            If `True`, the experimental version of the defined DMTF schema is
            installed.

            If `False`, (default) the final version of the defined
            DMTF schema is installed.

          verbose (:class:`py:bool`):
            If `True`, progress messages are output to stdout as the schema is
            downloaded and expanded. Default is `False`.

        Raises:
            ValueError: if the schema cannot be retrieved from the DMTF web
              site.
            TypeError:  if the `schema_version` is not a valid tuple with 3
              integer components

        """
        if not isinstance(schema_version, tuple):
            raise TypeError(
                _format("Schema_version {v!A} must be tuple",
                        v=schema_version))
        if len(schema_version) != 3:
            raise ValueError(
                _format("DMTF Schema must have 3 integer components, "
                        "(major, minor, update version) not {v!A}",
                        v=schema_version))  # is a tuple
        for i in schema_version:
            if not isinstance(i, six.integer_types):
                raise TypeError(
                    _format("{0!A} in schema_version {v!A} not integer",
                            i, v=schema_version))  # is a tuple
        schema_type = 'Experimental' if use_experimental else 'Final'
        self._schema_version = schema_version
        self._schema_root_dir = schema_root_dir

        mof_dir = 'mof{0}{1}'.format(schema_type, self.schema_version_str)
        self._schema_mof_dir = os.path.join(self._schema_root_dir, mof_dir)

        cim_schema_version = 'cim_schema_{0}'.format(self.schema_version_str)
        mof_zip_bn = '{0}{1}-MOFs.zip'.format(cim_schema_version, schema_type)
        self._schema_zip_url = \
            'https://www.dmtf.org/sites/default/files/cim/' \
            'cim_schema_v{0}{1}{2}/{3}'. \
            format(schema_version[0], schema_version[1], schema_version[2],
                   mof_zip_bn)
        schema_mof_bld_name = cim_schema_version + '.mof'

        self._schema_zip_file = os.path.join(self._schema_root_dir, mof_zip_bn)
        self._schema_pragma_file = os.path.join(self.schema_mof_dir,
                                                schema_mof_bld_name)

        self.verbose = verbose

        # Setup the local copy of the DMTF CIM schema which may include creating
        # the schema directory, downloading the schema zip file from the DMTF
        # web site, and extracting the schema into a subdirectory of the schema
        # directory.
        self._setup_dmtf_schema()

    @property
    def schema_version(self):
        """
        :func:`py:tuple`: The DMTF CIM schema version as a tuple of 3 integers
        with major version, minor version, and update version.

        Example: (2, 49, 0) defines DMTF CIM schema version 2.49.0.
        """
        return self._schema_version

    @property
    def schema_version_str(self):
        """
        :term:`string`: The DMTF CIM schema version as a string in the form
        ``major version>.<minor version>.<update version>``.

        Example: "2.49.0" defines DMTF CIM schema version 2.49.0.
        """
        return '{0}.{1}.{2}'.format(*self.schema_version)

    @property
    def schema_root_dir(self):
        """
        :term:`string`: Path name of the directory in which the DMTF CIM
        schema zip file is downloaded. The MOF files are extracted into the
        subdirectory indicated by :attr:`schema_mof_dir`.
        """
        return self._schema_root_dir

    @property
    def schema_mof_dir(self):
        """
        :term:`string`: Path name of the directory in which the DMTF CIM Schema
        MOF files are extracted from the downloaded zip file. This property can
        be used as the MOF compiler search path for compiling classes in the
        DMTF CIM schema. This directory will also contain the
        schema pragma file.
        """
        return self._schema_mof_dir

    @property
    def schema_pragma_file(self):
        """
        :term:`string`: Path name of the schema pragma file for the DMTF CIM
        schema. This file contains `#pragama include` statements for all of
        the classes and qualifier declarations of the schema.

        The classes are represented with one file per class, and the qualifier
        declarations are in the files `qualifiers.mof` and
        `qualifiers_optional.mof`.

        The path name of the schema pragma file is of the general form::

            <schema_mof_dir>/cim_schema_<schema_version_str>.mof

        Example::

            schemas/dmtf/moffinal2490/cim_schema_2.49.0.mof
        """
        return self._schema_pragma_file

    @property
    def schema_zip_file(self):
        """
        :term:`string`: Path name of the DMTF CIM schema zip file after being
        downloaded from the DMTF web site.
        """
        return self._schema_zip_file

    @property
    def schema_zip_url(self):
        """
        :term:`string`: URL of the DMTF CIM schema zip file that is
        downloaded from the DMTF web site.

        """
        return self._schema_zip_url

    def __str__(self):
        """
        Return a short string representation of the
        :class:`~pywbem_mock.DMTFCIMSchema` object for human consumption.
        This displays the major properties of the object.
        """

        return '{0}(schema_version_str={1}, schema_root_dir={2}, ' \
               'schema_mof_dir={3}, ' \
               'schema_pragma_file={4})'.format(
                   self.__class__.__name__,
                   self.schema_version_str,
                   self.schema_root_dir,
                   self.schema_mof_dir,
                   self.schema_pragma_file)

    def __repr__(self):
        """
        Return a string representation of the
        :class:`~pywbem_mock.DMTFCIMSchema` object that is suitable for
        debugging.
        """
        return '{0}(schema_version={1}, schema_root_dir={2}, ' \
               'schema_zip_file={3}, schema_mof_dir={4}, ' \
               'schema_pragma_file={5}, schema_zip_url={6})' \
               .format(self.__class__.__name__,
                       self.schema_version,
                       self.schema_root_dir,
                       self.schema_zip_file,
                       self.schema_mof_dir,
                       self.schema_pragma_file,
                       self.schema_zip_url)

    def _setup_dmtf_schema(self):
        """
        Install the DMTF CIM schema from the DMTF web site if it is not already
        installed. This includes downloading the DMTF CIM schema zip file from
        the DMTF web site and expanding that file into a subdirectory defined
        by `schema_mof_dir`.

        Once the schema zip file is downloaded into `schema_root_dir`, it is
        not re-downloaded if this function is recalled since DMTF CIM Schema
        releases are never modified; new update versions are released for minor
        changes. If the `schema_zip_file` is in the `schema_root_dir`
        directory, but no 'schema_mof_dir' subdirectory exists, the schema is
        unzipped.

        This allows the DMTF CIM schema zip file to be downloaded once and
        reused and the user to chose if they want to retain the extracted MOF
        files or remove them with :meth:`~pywbem_mock.DMTFCIMSchema.clean` when
        not being used.

        If the schema is to be committed a source repository such as git
        it is logical to commit only the DMTF CIM schema zip file. Creation of
        the `schema_mof_dir` subdirectory will be created when the
        :class:`pywbem_mock.DMTFCIMSchema` object is created.

        Raises:
            ValueError: If the schema cannot be retrieved from the DMTF web
              site.
            TypeError:  If the `schema_version` is not a valid tuple with 3
              integer components
        """

        def print_verbose(msg):
            """
            Inner method prints msg if self.verbose is `True`.
            """
            if self.verbose:
                print(msg)

        if not os.path.isdir(self.schema_root_dir):
            print_verbose(
                _format("Creating directory for CIM Schema archive: {0}",
                        self.schema_root_dir))
            os.mkdir(self.schema_root_dir)

        if not os.path.isfile(self.schema_zip_file):
            print_verbose(
                _format("Downloading CIM Schema archive from: {0}",
                        self.schema_zip_url))

            try:
                ufo = urlopen(self.schema_zip_url)
            except IOError as ie:
                os.rmdir(self.schema_root_dir)
                raise ValueError(
                    _format("DMTF Schema archive not found at url {0}: {1}",
                            self.schema_zip_url, ie))

            with open(self.schema_zip_file, 'wb') as fp:
                for data in ufo:
                    fp.write(data)

        if not os.path.isdir(self.schema_mof_dir):
            print_verbose(
                _format("Creating directory for CIM Schema MOF files: {0}",
                        self.schema_mof_dir))
            os.mkdir(self.schema_mof_dir)
        if not os.path.isfile(self._schema_pragma_file):
            print_verbose(
                _format("Unpacking CIM Schema archive: {0}",
                        self.schema_zip_file))

            zfp = None
            try:
                zfp = ZipFile(self.schema_zip_file, 'r')
                nlist = zfp.namelist()
                for file_ in nlist:
                    dfile = os.path.join(self.schema_mof_dir, file_)
                    if dfile[-1] == '/':
                        if not os.path.exists(dfile):
                            os.mkdir(dfile)
                    else:
                        with open(dfile, 'w+b') as dfp:
                            dfp.write(zfp.read(file_))
            finally:
                if zfp:
                    zfp.close()

    def build_schema_mof(self, class_names):
        """
        Build a string that includes the ``#include pragma`` statement for the
        CIM schema classes defined in `class_names` using the DMTF CIM
        schema defined by this instance of the class.

        The class names in `class_names` can be just leaf classes within the
        schema . The pywbem MOF compiler will search for dependent classes
        including superclasses, and other classes referenced as the classes are
        compiled.

        It builds a compilable MOF string in the form::

            #pragma locale ("en_US")
            #pragma include ("System/CIM_ComputerSystem.mof")
            ...

        with a ``#pragma include`` for each class name in the `class_names`
        list.

        The resulting set of #pragma statements is in the same order as the
        original pragmas.  In some cases that could be important because the
        cim schema pragma file is ordered to missing dependencies and conflicts
        between class mof compiles

        Parameters:

          class_names (:term:`py:list` of :term:`string` or :term:`string`):
            These must be class names of classes in the DMTF CIM schema
            represented by this :class:`~pywbem_mock.DMTFCIMSchema` object.
            This parameter can be a string if there is only a single class name
            to be included.

            If the returned string is compiled, the MOF compiler will search
            the directory defined by `schema_mof_dir` for classes referenced by
            each class in this list including super classes, classed defined by
            reference properties, and classes defined by an embedded-object.

        Returns:
            :term:`string`: Valid MOF containing pragma statements for
            all of the classes in `schema_classes`.

        Raises:
            ValueError: If any of the classnames in `schema_classes` are not in
              the DMTF CIM schema installed
        """
        return build_schema_mof(class_names, self.schema_pragma_file)

    def clean(self):
        """
        Remove all of the MOF files and the `schema_mof_dir` for the defined
        schema.  This is useful because while the downloaded schema is a single
        compressed zip file, it creates several thousand MOF files that take up
        considerable space.

        The next time the :class:`~pywbem_mock.DMTFCIMSchema` object for this
        `schema_version` and `schema_root_dir` is created, the MOF file are
        extracted again.
        """
        if os.path.isdir(self.schema_mof_dir):
            shutil.rmtree(self.schema_mof_dir)

    def remove(self):
        """
        The `schema_mof_dir` directory is removed if it esists and the
        `schema_zip_file` is removed if it exists. If the `schema_root_dir` is
        empty after these removals that directory is removed.
        """
        self.clean()
        if os.path.isfile(self.schema_zip_file):
            os.remove(self.schema_zip_file)
        if os.path.isdir(self.schema_root_dir) and \
                os.listdir(self.schema_root_dir) == []:
            os.rmdir(self.schema_root_dir)



"""
Pytest fixtures for pywbem end2end tests.
"""

from __future__ import absolute_import

import os
import warnings
import pytest
import yaml

from pywbem import WBEMServer, Error, ConnectionError, AuthError, \
    ToleratedServerIssueWarning  # pylint: disable=redefined-builtin

from .server_definition_file import ServerDefinition, \
    ServerDefinitionFile
from .utils import latest_profile_inst, ServerObjectCache, \
    WBEMConnectionAsserted, server_func_asserted
from .assertions import assert_association_a1, assert_association_a2, \
    assert_association_a3, assert_association_a4, assert_association_a5, \
    assert_association_a6

# Server nickname or server group nickname in WBEM server definition file
TESTSERVER = os.getenv('TESTSERVER', 'default')
SD_LIST = ServerDefinitionFile().list_servers(TESTSERVER)

# Profile definition file (in end2end directory).
PROFILES_YAML_FILE = os.path.join('tests', 'profiles', 'profiles.yml')

# Profile definition list.
# The list items are profile definition items, as described in the profile
# definition file.
with open(PROFILES_YAML_FILE, 'r') as _fp:
    PROFILE_DEFINITION_LIST = yaml.load(_fp)
del _fp  # pylint: disable=undefined-loop-variable


def fixtureid_server_definition(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `server_definition()`.

    Parameters:
      * fixture_value (ServerDefinition): The server definition of the
        WBEM server the test runs against.
    """
    sd = fixture_value
    assert isinstance(sd, ServerDefinition)
    return "server_definition={0}".format(sd.nickname)


@pytest.fixture(
    params=SD_LIST,
    scope='module',
    ids=fixtureid_server_definition
)
def server_definition(request):
    """
    Fixture representing the set of WBEM server definitions to use for the
    end2end tests.

    Returns the `ServerDefinition` object of each server to test against.
    """
    return request.param


@pytest.fixture(
    scope='module'
)
def wbem_connection(request, server_definition):
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Fixture representing the set of WBEMConnection objects to use for the
    end2end tests.  This fixture uses the
    implementation_namespace item from the server_definition as the
    default namespace if that item exists in the server_definition.

    Returns the `WBEMConnection` object of each server to test against.
    """
    sd = server_definition  # ServerDefinition object

    # Skip this server if we had a skip reason in an earlier attempt
    skip_msg = getattr(sd, 'skip_msg', None)
    if skip_msg:
        pytest.skip("Remembered skip reason from earlier attempt: {0}".
                    format(skip_msg))

    x509 = dict(cert_file=sd.cert_file, key_file=sd.key_file) \
        if sd.cert_file and sd.key_file else None

    # Define the WBEMConnection default_namespace as the
    # implementation_namespace if that data exists, else define it as
    # None which invokes root/cimv2 as the default namespace.
    # NOTE: Since WBEMConnection assigns root/cimv2 as default, there is
    # no way to really know if a valid implementation namespace was defined.
    conn = WBEMConnectionAsserted(
        sd.url, (sd.user, sd.password),
        x509=x509,
        ca_certs=sd.ca_certs,
        no_verification=sd.no_verification,
        default_namespace=sd.implementation_namespace,
        timeout=10,
        server_definition=sd)

    # Check that the server can be reached and authenticated with, by issuing
    # some quick operation. Operation failures are tolerated (e.g. GetQualifier
    # is not supported on SFCB), and the connection and authgentication errors
    # can still be detected because they will be detected before the operation
    # fails.
    try:
        conn.GetQualifier(
            'Association',
            asserted=False)
    except ConnectionError as exc:
        msg = "Server {0} at {1}: Server cannot be reached: {2} - {3}". \
            format(sd.nickname, sd.url, exc.__class__.__name__, exc)
        sd.skip_msg = msg
        warnings.warn(msg, ToleratedServerIssueWarning)
        pytest.skip(msg)
    except AuthError as exc:
        msg = "Server {0} at {1}: Server cannot be authenticated with: " \
            "{2} - {3}". \
            format(sd.nickname, sd.url, exc.__class__.__name__, exc)
        sd.skip_msg = msg
        warnings.warn(msg, ToleratedServerIssueWarning)
        pytest.skip(msg)
    except Error:
        # Any other error may be due to the WBEM server not supporting
        # qualifier operations. In that case, we don't optimize.
        sd.skip_msg = None
    else:
        sd.skip_msg = None

    return conn


def fixtureid_default_namespace(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `default_namespace()`.

    Parameters:
      * fixture_value (string): The default namespace for the test.
    """
    ns = fixture_value
    return "default_namespace={0}".format(ns)


@pytest.fixture(
    params=[None, 'root/cimv2', 'interop', 'root/interop'],
    scope='module',
    ids=fixtureid_default_namespace
)
def default_namespace(request):
    """
    Fixture representing the set of default namespaces to open connections
    with.

    Returns the default namespace as a string.
    """
    return request.param


def fixtureid_profile_definition(fixture_value):
    """
    Return a fixture ID to be used by py.test, for fixture
    `profile_definition()`.

    Parameters:
      * fixture_value (dict): The profile definition dictionary of the profile.
    """
    pd = fixture_value
    assert isinstance(pd, dict)
    assert 'registered_name' in pd
    assert 'registered_org' in pd
    return "profile_definition={0}:{1}:{2}". \
        format(pd['registered_org'], pd['registered_name'],
               pd.get('registered_version', 'any'))


def _apply_profile_definition_defaults(pd):
    """Set the profile definition defaults"""
    if 'reference_direction' not in pd:
        pd['reference_direction'] = None
    if 'registered_version' not in pd:
        pd['registered_version'] = None
    if pd['reference_direction'] is None:
        org = pd['registered_org']
        if org == 'SNIA':
            pd['reference_direction'] = 'snia'
        elif org == 'DMTF':
            pd['reference_direction'] = 'dmtf'
        else:
            raise ValueError(
                "Profile definition error: The reference_direction item of a "
                "profile definition can only be defaulted when the registered "
                "organisation is DMTF or SNIA, but it is {0}".format(org))
    if 'scoping_class' not in pd:
        pd['scoping_class'] = None
    if 'scoping_path' not in pd:
        pd['scoping_path'] = None
    return pd


def single_profile_definition(org, name, version=None):
    """
    Return a single profile definition dictionary for the specified org and
    name, and optionally version. If version is None, it matches the
    profile definition with the latest version. Profile definitions that do
    not specify a version always match.
    Optional items in the profile definition dictionary are generated with
    their default values.

    If a profile definition does not exist, returns None.
    """
    assert org is not None
    assert name is not None

    # Determine profile definitions with the specified org and name
    pd_list = []
    for pd in PROFILE_DEFINITION_LIST:
        if org != pd['registered_org'] or name != pd['registered_name']:
            continue
        pd_list.append(pd)

    # Determine profile definition with the specified version or latest
    if version:
        # Search profile definition with that version or without a version
        version_info = version.split('.')
        for pd in pd_list:
            pd_version = pd.get('registered_version', None)
            if not pd_version:
                # A profile definition without a version always matches
                return _apply_profile_definition_defaults(pd.copy())
            pd_version_info = pd_version.split('.')
            if pd_version_info == version_info:
                return _apply_profile_definition_defaults(pd.copy())
    else:
        # Search profile definition with latest version or without a version
        latest_pd_version_info = (-1, -1, -1)
        latest_pd = None
        for pd in pd_list:
            pd_version = pd.get('registered_version', None)
            if not pd_version:
                # A profile definition without a version always matches
                return _apply_profile_definition_defaults(pd.copy())
            pd_version_info = pd_version.split('.')
            if pd_version_info > latest_pd_version_info:
                latest_pd_version_info = pd_version_info
                latest_pd = pd
        if latest_pd:
            return _apply_profile_definition_defaults(latest_pd.copy())

    # No such profile definition exists
    return None


@pytest.fixture(
    params=PROFILE_DEFINITION_LIST,
    scope='module',
    ids=fixtureid_profile_definition
)
def profile_definition(request):
    """
    Fixture representing the set of profile definitions to use for the
    end2end tests.

    Returns the profile definition dictionary of each profile to test against.
    For dict items, see the description in profiles.yaml.

    Optional items in the dictionary are generated with their default values.
    """
    pd = _apply_profile_definition_defaults(request.param.copy())
    return pd


@pytest.fixture(
    params=[
        # the function for a1 must be first, because its results are stored
        # in the object cache for use by the subsequent functions.
        assert_association_a1,
        assert_association_a2,
        assert_association_a3,
        assert_association_a4,
        assert_association_a5,
        assert_association_a6,
    ],
    scope='module',
)
def assert_association_func(request):
    """
    Fixture representing a function for asserting association traversal.

    Returns the function.

    Interface of the function:

        Parameters:
          conn, profile_id, source_path, source_role, assoc_class, far_role,
          far_class

        Returns:
            tuple of:
            - list of associated (far end) instances
            - list of association instances
    """
    return request.param


class ProfileTest(object):
    """
    Base class for end2end tests on a specific profile.
    """

    object_cache = ServerObjectCache()

    def init_profile(self, conn, profile_org, profile_name,
                     profile_version=None):
        """
        Initialize attributes for the profile.

        The following instance attributes are set:
          * conn: conn argument (WBEMConnection).
          * profile_org: profile_org argument (registered org string).
          * profile_name: profile_name argument (registered name string).
          * profile_version: profile_version argument (registered version
            string), may be None.
          * server: WBEMServer object created from the connection.
          * profile_definition: Profile definition dictionary for the profile.
          * profile_id: org:name:version string to identify the profile.
            The version defaults to 'any' if profile_version is None.
          * profile_inst: CIMInstance for the CIM_RegisteredProfile object for
            the profile. If it does not exist, the testcase is skipped.
        """
        # pylint: disable=attribute-defined-outside-init
        assert conn is not None
        assert profile_org is not None
        assert profile_name is not None
        self.conn = conn
        self.profile_org = profile_org
        self.profile_name = profile_name
        self.profile_version = profile_version  # May be None

        self.server = WBEMServer(conn)
        self.profile_definition = single_profile_definition(
            self.profile_org, self.profile_name, self.profile_version)
        assert self.profile_definition is not None

        self.profile_id = "{0}:{1}:{2}".format(
            self.profile_org, self.profile_name, self.profile_version or 'any')
        profile_insts_id = self.profile_id + ':profile_insts'

        try:
            profile_insts = self.object_cache.get_list(
                conn.url, profile_insts_id)
        except KeyError:
            profile_insts = server_func_asserted(
                self.server, 'get_selected_profiles',
                registered_org=self.profile_org,
                registered_name=self.profile_name,
                registered_version=self.profile_version)
            self.object_cache.add_list(
                conn.url, profile_insts_id, profile_insts)

        if not profile_insts:
            pytest.skip("{0} {1} profile (version {2}) is not advertised "
                        "on server {3!r}".
                        format(self.profile_org, self.profile_name,
                               self.profile_version or 'any', self.conn.url))

        self.profile_inst = latest_profile_inst(profile_insts)
        # pylint: enable=attribute-defined-outside-init

    def init_central_instances(self):
        """
        Initialize the central_inst_paths attribute with the central instances
        of the profile. init_profile() must have been called before.

        The following instance attributes are set:
          * central_inst_paths: central instances as list of CIMInstanceName.
        """

        # init_profile() must have been called before.
        assert getattr(self, 'profile_org', None)
        assert getattr(self, 'profile_name', None)

        central_paths_id = self.profile_id + ':central_paths'

        try:
            central_paths = self.object_cache.get_list(
                self.conn.url, central_paths_id)
        except KeyError:
            central_paths = server_func_asserted(
                self.server, 'get_central_instances',
                profile_path=self.profile_inst.path,
                central_class=self.profile_definition['central_class'],
                scoping_class=self.profile_definition['scoping_class'],
                scoping_path=self.profile_definition['scoping_path'],
                reference_direction=self.profile_definition[
                    'reference_direction'])
            self.object_cache.add_list(
                self.conn.url, central_paths_id, central_paths)

        # pylint: disable=attribute-defined-outside-init
        self.central_inst_paths = central_paths
